from lantz import Feat, DictFeat, Action
from lantz.messagebased import MessageBasedDriver
from pyvisa import constants

from time import sleep


class CoBriteDX1(MessageBasedDriver):
    """

    """

    # Set these parameters from the datasheet of your model.
    channels = ['1,1,1']  # can add additional channels here!

    freq_min = 191.3  # THz
    freq_max = 196.25  # THz

    min_power = 6  # dBm
    max_power = 15.5  # dBm
    offset_lim = 12.0  # GHz

    sbs_lims = (0, 1, 0.1)

    THz_to_nm = 299792  # 1 THz in nm
    lambda_min = THz_to_nm/freq_max
    lambda_max = THz_to_nm/freq_min

    on_off = {'off': 0, 'on': 1}
    yes_no = {'no': 0, 'yes': 1}

    DEFAULTS = {'ASRL': {'write_termination': ';',
                         'read_termination': ';',
                         'baud_rate': 115200,
                         'data_bits': 8,
                         'parity': constants.Parity.none,
                         'stop_bits': constants.StopBits.one,
                         'encoding': 'utf-8',
                         'timeout': 2000}}

    def initialize(self):
        """
        """
        super().initialize()
        sleep(1)
        self.echo = 'off'  # need to do this to drop command echoing

    @Feat(values=on_off)
    def echo(self):
        """
        Checks to see if ECHO mode is enabled. Note that the code in this
        driver assumes that the command echo is disabled.
        """
        return int(self.query(':SYS:ECHO?'))

    @echo.setter
    def echo(self, status):
        """
        Sets echo mode to be controlled by on_off.
        """
        print(':SYS:ECHO {}'.format(status))
        return self.query(':SYS:ECHO {}'.format(status))

    @Feat()
    def idn(self):
        """
        Returns the laser firmware and hardware information.
        """
        return self.query('*idn?')

    @Feat()
    def layout(self):
        """
        Returns the configuration of the attached system.

        This will be in format:
        SYSTEM <Master chassis type>, <slave chassis type>,
        <slave chassis type>,...\n <Slot address>, <card type>, \n
        <Slot address>, <card type>\n...
        """
        return self.query(':SYS:LAY?')

    @DictFeat(keys=channels, limits=(lambda_min, lambda_max))
    def wavelength(self, channel):
        """
        Returns the current laser wavelength, in nm.
        """
        result = self.query(':SOUR:WAV? {}'.format(channel))
        return float(result.replace(channel + ',', ''))

    @wavelength.setter
    def wavelength(self, channel, lambd):
        """
        Sets the current laser wavelength to lambda (nm).
        """
        delay = 1
        value = self.query(':SOUR:WAV {}, {:.4f}'.format(channel, lambd))
        while not self.operation_complete:
            sleep(delay)
        #self.read()
        return value

    @DictFeat(keys=channels, limits=(lambda_min, lambda_max))
    def wavelength_lim(self, channel):
        """
        Returns wavelength limits of the laser at channel in nm.
        """
        result = self.query(':SOUR:WAV:LIM? {}'.format(channel))
        return self.split_floats(channel, result)

    @DictFeat(keys=channels, limits=(freq_min, freq_max))
    def frequency(self, channel):
        """
        Returns the frequency of the laser at channel in THz.
        """
        return float(self.query(':SOUR:FREQ? {}'.format(channel)))

    @frequency.setter
    def frequency(self, channel, THz):
        """
        Sets the frequency of the laser at channel to THz.
        """
        delay = 1
        value = self.write(':SOUR:FREQ {},{:.4f}'.format(channel, THz))
        while not self.operation_complete:
            sleep(delay)
        self.read()
        return value

    @DictFeat(keys=channels, limits=(freq_min, freq_max))
    def frequency_lim(self, channel):
        """
        Return the frequency limits of the laser at channel in THz.
        """
        result = self.query(':SOUR:FREQ:LIM? {}'.format(channel))
        return self.split_floats(channel, result)

    @DictFeat(keys=channels, limits=(0, offset_lim))
    def offset(self, channel):
        """
        Returns the offset of the laser at channel in GHz.
        """
        return float(self.query(':SOUR:OFF? {}'.format(channel)))

    @offset.setter
    def offset(self, channel, GHz):
        """
        Sets the offset of the laser at channel to GHz.
        """
        delay = 1
        value = self.write(':SOUR:OFF {},{:.4f}'.format(channel, GHz))
        while not self.operation_complete:
            sleep(delay)
        self.read()
        return value

    @DictFeat(keys=channels, limits=(0, offset_lim))
    def offset_lim(self, channel):
        """
        Returns the fine tuning limits of the laser at channel in GHz.
        """
        return float(self.query(':SOUR:OFF:LIM? {}'.format(channel)))

    @DictFeat(keys=channels, limits=(min_power, max_power))
    def power(self, channel):
        """
        Returns the laser power.
        """
        return float(self.query(':SOUR:POW? {}'.format(channel)))

    @power.setter
    def power(self, channel, dBm):
        """
        Sets the power of the laser at channel to dBm.
        """
        delay = 1
        value = self.write(':SOUR:POW {},{:.4f}'.format(channel, dBm))
        while not self.operation_complete:
            sleep(delay)
        self.read()
        return value

    @DictFeat(keys=channels, limits=(0, max_power))
    def power_lim(self, channel):
        """
        Returns maximum power of the laser at channel in dBm.
        """
        result = self.query(':SOUR:FREQ:LIM? {}'.format(channel))
        return self.split_floats(channel, result)

    @DictFeat(keys=channels)
    def limits(self, channel):
        """
        Returns all limits of the laser at channel in form:
        [Min freq (THz), Max freq(THz), Tune range (GHz), Min pow (dBm),
        Max power (dBm)]
        """
        result = self.query(':SOUR:LIM? {}'.format(channel))
        return self.split_floats(channel, result)

    @DictFeat(keys=channels, values=on_off)
    def state(self, channel):
        """
        Returns the current state of the laser at channel.
        """
        return int(self.query(':SOUR:STAT? {}'.format(channel)))

    @state.setter
    def state(self, channel, on_status):
        """
        Sets the state of the laser at channel to on_status
        """
        delay = 1
        value = self.write(':SOUR:STAT {},{}'.format(channel, on_status))
        while not self.operation_complete:
            sleep(delay)
        self.read()
        return value

    @DictFeat(keys=channels)
    def configuration(self, channel):
        """
        Returns the current laser configuration in format:
        [Freq (THz), Fine tuning (GHz), Power (dBm), On/off (0/1), Busy (0/1),
        Dither (0/1)]
        """
        result = self.query(':SOUR:CONF? {}'.format(channel)).split(',')
        return (list(map(float, result[0:3])) + list(map(int, result[3:6])))

    @DictFeat(keys=channels, values=yes_no)
    def busy(self, channel):
        """
        Query state of a device.
        If 1, the laser is currently tuning and not settled.
        """
        return int(self.query(':SOUR:BUSY? {}'.format(channel)))

    @DictFeat(keys=channels)
    def monitor(self, channel):
        """
        Returns monitor information from laser in format:
        [LD chip temp (deg C), LD base temp (deg C), LD chip current (mA),
        TEC current (mA)]
        """
        result = self.query(':SOUR:MON? {}'.format(channel))
        return self.split_floats(channel, result)

    @DictFeat(keys=channels, values=on_off)
    def dither(self, channel):
        """
        Queries whether or not the laser at channel is on or off.
        """
        return int(self.query(':SOUR:DIT? {}'.format(channel)))

    @dither.setter
    def dither(self, channel, dither_status):
        """
        Sets the dither tone of the laser at channel to be dither_status.
        """
        return self.query(':SOUR:DIT {},{}'.format(channel, dither_status))

    @DictFeat(keys=channels, values=on_off)
    def sbs(self, channel):
        """
        Queries whether or not the stimulated Brillouin scattering (SBS)
        supression is on or off.
        """
        return int(self.query(':SOUR:SBS:STAT? {}'.format(channel)))

    @sbs.setter
    def sbs(self, channel, sbs_status):
        """
        Sets the stimulated Brillouin scattering (SBS) supression to be
        sbs_status.
        """
        return self.query(':SOUR:SBS:STAT {},{}'.format(channel, sbs_status))

    @Feat(limits=sbs_lims)
    def sbs_freq(self):
        """
        Queries the stimulated Brillouin scattering (SBS) supression amplitude
        for all lasers in GHz.
        """
        return float(self.query(':SOUR:SBS:FREQ?'))

    @sbs_freq.setter
    def sbs_freq(self, GHz):
        """
        Sets the stimulated Brillouin scattering (SBS) supression amplitude
        for all lasers to GHz.
        """
        return self.query(':SOUR:SBS:FREQ {}'.format(GHz))

    @Feat()
    def operation_complete(self):
        """
        Operation complete query, checks to see if all lasers are settled.
        """
        if self.query('*opc?') == '1':
            return True
        else:
            return False

    @Action()
    def save_curr_state(self, channel):
        """
        Permanently saves the current laser port state for channel, which will
        be loaded again after a power on/off cycle or reset. Autostart must be
        enabled for this to work.
        """
        if channel in self.channels:
            return self.query(':SYS:SCSTAT'.format(channel))
        else:
            print('Invalid channel.')
            return 0

    @Feat(values=on_off)
    def autostart(self):
        """
        Check to see if laser autostart is enabled.
        """
        return int(self.query(':SYS:AUTOSTA?'))

    @autostart.setter
    def autostart(self, autostart_status):
        """
        Sets the laser autostart to be autostart_status.
        """
        return self.query(':SYS:AUTOSTA'.format(autostart_status))

    def split_floats(self, channel, result):
        """
        Helper function for returning tuples returned during communication.
        """
        ret_vals = []
        for val in result.replace(channel + ',', '').split(','):
            ret_vals.append(float(val))
        return ret_vals

if __name__ == '__main__':
    with CoBriteDX1.via_serial(8) as inst:
        inst.echo = 'off'
        print('== System Information ==')
        print('IDN:{}'.format(inst.idn))
        print('Layout:{}'.format(inst.layout))
        print('Autostart:{}'.format(inst.autostart))

        print('== System Control Demo ==')

        for chan in inst.channels:
            print('Laser at channel: {}'.format(chan))
            print('== Laser Parameters ==')
            print('Wavelength: {} nm'.format(inst.wavelength[chan]))
            print('Frequency: {} THz'.format(inst.frequency[chan]))
            print('Power: {} dBm'.format(inst.power[chan]))
            print('Offset: {} GHz'.format(inst.offset[chan]))
            print('State: {}'.format(inst.state[chan]))
            print('Busy?: {}'.format(inst.busy[chan]))
            print('Configuration: {}'.format(inst.configuration[chan]))
            print('Dither?: {}'.format(inst.dither[chan]))
            print('SBS?: {}'.format(inst.sbs[chan]))
            print('SBS Frequency: {} GHz'.format(inst.sbs_freq))

            print('== Laser limits ==')
            print('Wavelength limits: {} nm'.format(inst.wavelength_lim[chan]))
            print('Frequency limits: {} THz'.format(inst.frequency_lim[chan]))
            print('Offset limit: {} GHz'.format(inst.offset_lim[chan]))
            print('Power limit: {} dBm'.format(inst.power_lim[chan]))
            print('All limits: {}'.format(inst.limits[chan]))

            print('== Laser Control Demo ==')
            #print('Saving current state...')
            #inst.save_curr_state(channel=chan)

            inst.frequency[chan] = 195.3
            print('Configuration: {}'.format(inst.configuration[chan]))
            inst.wavelength[chan] = 1530.1234
            print('Configuration: {}'.format(inst.configuration[chan]))
            inst.frequency[chan] = 191.3
            inst.offset[chan] = 6.5
            inst.state[chan] = 'on'
            print('Configuration: {}'.format(inst.configuration[chan]))
            print('Monitor: {}'.format(inst.monitor[chan]))
            inst.offset[chan] = 12
            inst.power[chan] = 6.1
            print('Configuration: {}'.format(inst.configuration[chan]))
            print('Busy:{}'.format(inst.busy[chan]))
            inst.power[chan] = 15.5
            print('Configuration: {}'.format(inst.configuration[chan]))
            inst.offset[chan] = 0.0
            print('Configuration: {}'.format(inst.configuration[chan]))
            inst.state[chan] = 'off'
            print('Configuration: {}'.format(inst.configuration[chan]))



# -*- coding: utf-8 -*-
"""
    lantz.drivers.lakeshore
    ~~~~~~~~~~~~~~~~~~~~~

    :company: Lakeshore Cryotronics
    :description: Measurement and control equipment
    :website: http://www.lakeshore.com/

"""

from .lakeshore332 import Lakeshore332
from .lakeshore475 import Lakeshore475

__all__ = ['Lakeshore332', 'Lakeshore475']



from lantz import Driver, Action, Feat, DictFeat, ureg
import requests
import re

class WebPowerSwitch7(Driver):
	timeout = 1

	def __init__(self, host='192.168.1.20:80', user='admin', pwd='1234', *args, **kwargs):
		self.host, self.user, self.pwd = host, user, pwd
		super().__init__(*args, **kwargs)

	def send_cmd(self, cmd):
		url = "http://%s/%s" % (self.host, cmd)
		resp = requests.get(url, auth=(self.user, self.pwd,),  timeout=self.timeout)
		return resp

	@Action()
	def get_status(self):
		resp = self.send_cmd('status')
		pattern = lambda id: '<div id="'+id+'">(..)</div>'
		to_bin = lambda string: bin(int(string, 16))[2:].zfill(8)
		to_bool = lambda bin_array: list(map(lambda x: x=='1', bin_array))
		def get_val(id):
			bool_arr = to_bool(to_bin(re.search(pattern(id), str(resp.content)).group(1)))
			bool_arr.reverse()
			return bool_arr
		
		return {k:get_val(k) for k in ['state', 'lock', 'perm']}

	@DictFeat(keys=range(1,9), values={True: 'ON', False: 'OFF'})
	def state(self, key):
		return 'ON' if self.get_status()['state'][key-1] else 'OFF'

	@state.setter
	def state(self, key, value):
		self.send_cmd('outlet?{}={}'.format(key, value))



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.stanford
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :company: Standford Research Systems.
    :description: Manufactures test instruments for research and industrial applications
    :website: http://www.thinksrs.com/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .sr830 import SR830Serial, SR830GPIB

__all__ = ['SR830Serial', 'SR830GPIB']



# -*- coding: utf-8 -*-

from lantz.messagebased import MessageBasedDriver
from lantz import Feat

class PM100D(MessageBasedDriver):

    DEFAULTS = {
        'COMMON': {
            'read_termination': '\n',
            'write_termination': '\n',
        },
    }

    @Feat(read_once=True)
    def idn(self):
        return self.query('*IDN?')

    @Feat(units='W')
    def power(self):
        return float(self.query('MEAS:POWER?'))

    @Feat(units='nm')
    def correction_wavelength(self):
        return float(self.query('SENSE:CORRECTION:WAVELENGTH?'))

    @correction_wavelength.setter
    def correction_wavelength(self, wavelength):
        self.write('SENSE:CORRECTION:WAVELENGTH {}'.format(wavelength))

    @Feat()
    def correction_wavelength_range(self):
        cmd = 'SENSE:CORRECTION:WAVELENGTH? {}'
        cmd_vals = ['MIN', 'MAX']
        return tuple(float(self.query(cmd.format(cmd_val))) for cmd_val in cmd_vals)



# -*- coding: utf-8 -*-
"""
    lantz.drivers.agilent
    ~~~~~~~~~~~~~~~~~~~~~~
    :company: Agilent Technologies.
    :description: Manufactures test instruments for research and industrial applications
    :website: http://www.agilent.com/home
    ----
    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from .n51xx import N51xx
from .ag33220A import Ag33220A
from .ag81130a import Ag81130A
from .e8257c import E8257C
from .AG33522a import AG33522A

__all__ = ['N51xx', 'Ag33220A', 'Ag81130A', 'AG33522A', 'E8257C']



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.mpb.vfl
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz import Action, Feat
from lantz.drivers.legacy.serial import SerialDriver


class VFL(SerialDriver):
    """Driver for any VFL MPB Communications laser.
    """

    ENCODING = 'ascii'

    RECV_TERMINATION = '\rD >'
    SEND_TERMINATION = '\r'

    BAUDRATE = 9600
    BYTESIZE = 8
    PARITY = 'none'
    STOPBITS = 1

    #: flow control flags
    RTSCTS = False
    DSRDTR = False
    XONXOFF = False

    @Feat(read_once=True)
    def idn(self):
        """Identification of the device
        """
        return self.query('GETMODEL')

    @Feat()
    def status(self):
        """Current device status
        """
        ans = self.query('shlaser')
        return ans.split('\r')

    # ENABLE LASER
    @Feat(values={True: '1', False: '0'})
    def enabled(self):
        """Method for turning on the laser
        """
        return self.query('GETLDENABLE')

    @enabled.setter
    def enabled(self, value):
        self.query('SETLDENABLE ' + value)

    # LASER'S CONTROL MODE AND SET POINT

    @Feat(values={'APC': '1', 'ACC': '0'})
    def ctl_mode(self):
        """To handle laser diode current (mA) in Active Current Control Mode
        """
        return self.query('GETPOWERENABLE')

    @ctl_mode.setter
    def ctl_mode(self, value):
        self.query('POWERENABLE {}'.format(value))

    @Feat(units='mA')
    def current_sp(self):
        """To handle laser diode current (mA) in Active Current Control Mode
        """
        return float(self.query('GETLDCUR 1'))

    @current_sp.setter
    def current_sp(self, value):
        self.query('SETLDCUR 1 {:.1f}'.format(value))

    @Feat(units='mW')
    def power_sp(self):
        """To handle output power set point (mW) in APC Mode
        """
        return float(self.query('GETPOWER 0'))

    @power_sp.setter
    def power_sp(self, value):
        self.query('SETPOWER 0 {:.0f}'.format(value))

    # LASER'S CURRENT STATUS

    @Feat(units='mW')
    def power(self):
        """To get the laser emission power (mW)
        """
        return float(self.query('POWER 0'))

    @Feat(units='mA')
    def ld_current(self):
        """To get the laser diode current (mA)
        """
        return float(self.query('LDCURRENT 1'))

    @Feat(units='degC')
    def ld_temp(self):
        """To get the laser diode temperature (ºC)
        """
        return float(self.query('LDTEMP 1'))

    @Feat(units='mA')
    def tec_current(self):
        """To get the thermoelectric cooler (TEC) current (mA)
        """
        return float(self.query('TECCURRENT 1'))

    @Feat(units='degC')
    def tec_temp(self):
        """To get the thermoelectric cooler (TEC) temperature (ºC)
        """
        return float(self.query('TECTEMP 1'))

    # SECOND HARMONIC GENERATOR METHODS

    @Feat(units='degC')
    def shg_temp_sp(self):
        """To handle the SHG temperature set point
        """
        return float(self.query('GETSHGTEMP'))

    @shg_temp_sp.setter
    def shg_temp_sp(self, value):
        self.query('GETSHGTEMP {:.2f}'.format(value))

    @Feat(units='degC')
    def shg_temp(self):
        """To get the SHG temperature
        """
        return float(self.query('SHGTEMP'))

    @Feat()
    def shg_tune_info(self):
        """Getting information about laser ready for SHG tuning
        """
        info = self.query('GETSHGTUNERDY').split()
        if info[0] == '0':
            ready = 'Laser not ready for SHG tuning. '
        else:
            ready = 'Laser ready for SHG tuning. '

        schedule = 'Next SHG tuning scheduled in {} '.format(info[1])
        schedule += 'hours of operation. '
        warm = 'Warm-up period expires in {} seconds.'.format(info[2])

        ans = ready + schedule + warm
        return ans

    @Feat()
    def shg_tuning(self):
        """Initiating SHG tuning
        """
        state = self.query('GETSHGTUNESTATE').split()
        tuning = error = ''

        if state[0] == '0':
            tuning = 'No SHG tuning performed since last reset. '
        elif state[0] == '3':
            tuning = 'SHG tuning in progress. '
        elif state[0] == '1':
            tuning = 'SHG tuning completed successfully. '
        elif state[0] == '2':
            tuning = 'SHG tuning aborted. '

        if state[1] == '0':
            error = 'No error detected.'
        elif state[1] == '1':
            error = 'Error: Laser not running in APC.'
        elif state[1] == '8':
            error = 'Error: Output Power not stabilized.'

        return tuning + error

    @Action()
    def tune_shg(self):
        self.query('SETSHGCMD 1')

    @Action()
    def tune_shg_stop(self):
        self.query('SETSHGCMD 2')

if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')
    parser.add_argument('-p', '--port', type=str, default='COM3',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_screen(lantz.log.DEBUG)
    with VFL(args.port) as inst:
        if args.interactive:
            from lantz.ui.qtwidgets import start_test_app
            start_test_app(inst)
        else:
            # Add your test code here
            print('Non interactive mode')
            print(inst.idn)
            print(inst.shg_tuning)



# -*- coding: utf-8 -*-
"""
    lantz.drivers.newport.agilis
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of Agilis UC8/UC2 piezo controller

    Authors: Kevin Miao, Alexandre Bourassa
    Date: 12/14/2015

"""

from lantz import Action, Feat, DictFeat
from lantz.messagebased import MessageBasedDriver

from pyvisa.constants import Parity, StopBits

import time

class Agilis(MessageBasedDriver):

    comm_delay = 0.2

    DEFAULTS = {
        'ASRL': {
            'write_termination': '\r\n',
            'read_termination': '\r\n',
            'baud_rate': 921600,
            'parity': Parity.none,
            'stop_bits': StopBits.one,
            'timeout': None,
        }
    }

    # We need a synchronous decorator (or something to slow down function calls)
    # most commands need ~1s to complete, while select commands may need up to
    # 2 minutes
    def write(self, *args, **kwargs):
        super(Agilis, self).write(*args, **kwargs)
        time.sleep(self.comm_delay)
        return

    @Action()
    def reset(self):
        """
        Reset controller
        """
        self.write('RS')
        return

    #@Feat(values={1, 2, 3, 4})
    @Feat()
    def channel(self):
        retval = self.query('CC?')
        return int(retval.lstrip('CC'))

    @channel.setter
    def channel(self, value):
        """
        Set channel
        """
        self.write('CC{}'.format(value))

    @Feat()
    def delay(self, axis):
        return self.query('{}DL?'.format(axis))

    @delay.setter
    def delay(self, axis, value):
        self.write('{}DL{}'.format(axis, value))

    @Feat()
    def jog(self, axis):
        return self.query('{}JA?'.format(axis))

    @Action()
    def jog(self, axis, value):
        self.write('{}JA{}'.format(axis, value))

    @Action()
    def measure_position(self, axis):
        return self.query('{}MA'.format(axis))

    @Action()
    def move_to_limit(self, axis, value):
        self.write('{}MV{}'.format(axis, value))

    @Action()
    def move_to_limit(self, axis, value):
        while not self.limit_status[axis]:
            self.jog(axis, value)
        self.jog(axis, 0)

    @Action()
    def relative_move(self, axis, steps):
        self.write('{}PR{}'.format(axis, steps))

    @DictFeat()
    def step_amplitude(self, axis):
        val1 = self.query('{}SU-?'.format(axis))
        val2 = self.query('{}SU+?'.format(axis))
        stripped = '{}SU'.format(axis)
        return (val1.lstrip(stripped), val2.lstrip(stripped))


    @step_amplitude.setter
    def step_amplitude(self, axis, amplitude):
        if amplitude < 0:
            sign = '-'
        elif amplitude > 0:
            sign = '+'
        else:
            raise ValueError('amplitude cannot be 0')
        self.write('{}SU{}{:d}'.format(axis, sign, abs(amplitude)))

    @Action()
    def move(self, axis, value, mode):
        raise NotImplementedError

    @Feat()
    def limit_status(self):
        """
        PH0 if neither limit reached
        PH1 if only axis 1
        PH2 if only axis 2
        PH3 if both
        """
        retval = self.query('PH')
        self.log_debug(retval)
        code = int(retval.lstrip('PH'))
        status = {1: False, 2: False}
        status[1] = True if code & 1 else False
        status[2] = True if (code >> 1) & 1 else False
        return status

    @DictFeat()
    def status(self, axis):
        return self.query('{}TS'.format(axis))

    @Action()
    def stop(self, axis):
        self.write('{}ST'.format(axis))

    @Action()
    def zero(self, axis):
        self.write('{}ZP'.format(axis))

    @DictFeat()
    def steps(self, axis):
        retval = self.query('{}TP?'.format(axis))
        return int(retval.lstrip('{}TP'.format(axis)))

    @Feat()
    def version(self):
        return self.query('VE').strip()

    @Action()
    def calibrate(self, axis):
        self.zero(axis)
        self.step_amplitude[axis] = 50
        if self.limit_status[axis]:
            self.relative_move(axis, 100)
        if self.limit_status[axis]:
            self.relative_move(axis, -100)
        self.move_to_limit(axis, -3)
        self.step_amplitude[axis] = 50
        self.zero(axis)
        self.relative_move(axis, 100)
        self.move_to_limit(axis, 3)
        return self.steps[axis]

def main():
    import logging
    import sys
    from lantz.log import log_to_screen
    log_to_screen(logging.CRITICAL)
    res_name = sys.argv[1]
    fmt_str = "{:<20}|{:>20}"
    axis = 1
    with Agilis(res_name) as inst:
        inst.channel = 1
        print(fmt_str.format("Device version", inst.version))
        print(fmt_str.format("Current channel", inst.channel))
        amplitudes = inst.step_amplitude[axis]
        print(fmt_str.format("- Step amplitude", amplitudes[0]))
        print(fmt_str.format("+ Step amplitude", amplitudes[1]))
        print(inst.calibrate(1))
        print(inst.calibrate(2))
        inst.channel = 2
        print(inst.calibrate(1))
        print(inst.calibrate(2))

if __name__ == '__main__':
    main()



from lantz.messagebased import MessageBasedDriver
from lantz import Feat, DictFeat, Action

#from lantz import Q_

import socket
import warnings

class Ag33220A(MessageBasedDriver):
    """
    Lantz driver for interfacing with Aiglent 33220A function generator.

    Includes testing code, which should work out of the box assuming you give
    it the correct IP address.

    Author: P. Mintun
    Date: 10/21/2016
    Version: 0.1
    """

    DEFAULTS = {
        'COMMON': {
            'write_termination': '\n',
            'read_termination': '\n',
        }
    }

    function_dict = {
        'sine': 'SIN',
        'square': 'SQU',
        'ramp': 'RAMP',
        'pulse': 'PULS',
        'noise': 'NOIS',
        'DC': 'DC',
        'user': 'USER'
    }

    @Feat()
    def idn(self):
        """
        Identifiies the instrument.
        """
        return self.query('*IDN?')

    @Feat(values=function_dict)
    def function(self):
        """
        Returns the selected output function.
        """
        return self.query('FUNC?')

    @function.setter
    def function(self, func_type):
        """
        Sets the output function.
        """
        return self.write('FUNC {}'.format(func_type))

    @Feat()
    def frequency(self):
        """
        Queries the output frequency of the function generator.
        """
        return float(self.query('FREQ?'))

    @frequency.setter
    def frequency(self, Hz):
        """
        Sets the output frequency of the function generator (in Hz).
        """
        return self.write('FREQ {}'.format(Hz))

    @Feat(values={'on': 1, 'off': 0})
    def output_toggle(self):
         """
         Returns whether or not the output is on or off.
         """
         return int(self.query('OUTP?'))

    @output_toggle.setter
    def output_toggle(self, on_off):
        """
        Sets the output to be either on, or off.
        """
        return self.write('OUTP {}'.format(on_off))

    @Feat(limits=(0.01,10))
    def voltage(self):
        """
        Returns the amplitude voltage setting.
        """
        return float(self.query('VOLT?'))

    @voltage.setter
    def voltage(self, volts):
        """
        Sets the amplitude voltage setting to volts.
        """
        return self.write('VOLT {}'.format(volts))

    @Feat()
    def voltage_offset(self):
        """
        Returns the DC offset voltage, in volts.
        """
        return float(self.query('VOLT:OFFS?'))

    @voltage_offset.setter
    def voltage_offset(self, V_dc):
        """
        Sets the DC offset voltage to be V_dc, in volts.
        """
        return self.write('VOLT:OFFS {}'.format(V_dc))

    # special commands for pulse mode
    @Feat()
    def pulse_width(self):
        """
        Returns the pulse width in nanoseconds.
        """
        return float(self.query('FUNC:PULS:WIDT?'))

    @pulse_width.setter
    def pulse_width(self, sec):
        """
        Sets the pulse width in seconds.
        """
        return self.write('FUNC:PULS:WIDT {}'.format(sec))

    # special functions for square waves
    @Feat()
    def square_duty_cycle(self):
        """
        Returns the duty cycle of a square wave output, in percent.
        """
        return float(self.query('FUNC:SQU:DCYC?'))

    @square_duty_cycle.setter
    def square_duty_cycle(self, duty_cycle):
        """
        Sets the square wave duty cycle to be duty_cycle.
        """
        return self.write('FUNC:SQU:DCYC {}'.format(duty_cycle))


if __name__ == '__main__':
    address = 'xxx.yyy.zzz.aaa'
    inst_num = 'instYY'

    print('Need to set IP address and inst_num!')

    with Ag33220A('TCPIP0::{}::{}::INSTR'.format(address, inst_num)) as inst:
        print('Identification: {}'.format(inst.idn))

        print('Function: {}'.format(inst.function))
        inst.function = 'sine'
        print('Function: {}'.format(inst.function))
        inst.function = 'pulse'
        print('Function: {}'.format(inst.function))
        inst.function = 'square'
        print('Function: {}'.format(inst.function))


        print('Frequency:{}Hz'.format(inst.frequency))
        inst.frequency = 1000.0
        print('Frequency:{}Hz'.format(inst.frequency))
        inst.frequency = 20000.0
        print('Frequency:{}Hz'.format(inst.frequency))


        print('Output:{}'.format(inst.output_toggle))
        inst.output_toggle = 'off'
        print('Output:{}'.format(inst.output_toggle))
        inst.output_toggle = 'on'
        print('Output:{}'.format(inst.output_toggle))

        print('Amplitude voltage:{}V'.format(inst.voltage))
        inst.voltage = 2.5
        print('Amplitude voltage:{}V'.format(inst.voltage))
        inst.voltage = 5.0
        print('Amplitude voltage:{}V'.format(inst.voltage))

        print('Offset voltage:{}'.format(inst.voltage_offset))
        inst.voltage_offset = 2.1
        print('Offset voltage:{}'.format(inst.voltage_offset))
        inst.voltage_offset = 2.5
        print('Offset voltage:{}'.format(inst.voltage_offset))

        inst.function = 'pulse'
        print('Pulse width:{}s'.format(inst.pulse_width))
        inst.pulse_width = 50e-9
        print('Pulse width:{}s'.format(inst.pulse_width))
        inst.pulse_width = 20e-9
        print('Pulse width:{}s'.format(inst.pulse_width))

        inst.function = 'square'
        print('Duty cycle:{}'.format(inst.square_duty_cycle))
        inst.square_duty_cycle = 25.0
        print('Duty cycle:{}'.format(inst.square_duty_cycle))
        inst.square_duty_cycle = 50.0
        print('Duty cycle:{}'.format(inst.square_duty_cycle))



# -*- coding: utf-8 -*-
"""
    lantz.drivers.newport.motion axis
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    General class that implements the commands used for several smaract motion
    drivers using the ASCII mode (via serial or serial via USB).

    :copyright: 2018, see AUTHORS for more details.
    :license: GPL, see LICENSE for more details.

    Source: Instruction Manual (Newport)

"""


from lantz.feat import Feat
from lantz.action import Action
from lantz.messagebased import MessageBasedDriver
from pyvisa import constants
from lantz import Q_, ureg
from lantz.processors import convert_to
from lantz.drivers.motion import MotionAxisMultiple, MotionControllerMultiAxis, BacklashMixing
import time
import numpy as np

formats = {'one_param': ''}

class SCU(MessageBasedDriver, MotionControllerMultiAxis):
    """ Driver for SCU controller with multiple axis

    """
    DEFAULTS = {
                'COMMON': {'write_termination': '\n',
                           'read_termination': '\n', },
                'ASRL': {
                    'timeout': 100,  # ms
                    'encoding': 'ascii',
                    'data_bits': 8,
                    'baud_rate': 9600,
                    'parity': constants.Parity.none,
                    'stop_bits': constants.StopBits.one,
                    #'flow_control': constants.VI_ASRL_FLOW_NONE,
                    'flow_control': constants.VI_ASRL_FLOW_XON_XOFF,  # constants.VI_ASRL_FLOW_NONE,
                    },
                }

    def initialize(self):
        super().initialize()
        self.detect_axis()

    def query(self, command, *, send_args=(None, None), recv_args=(None, None)):
        return MotionControllerMultiAxis.query(self, ':{}'.format(command),
                                 send_args=send_args, recv_args=recv_args)

    def write(self, command, *args, **kwargs):
        return MotionControllerMultiAxis.write(self,':{}'.format(command),
                                 *args, **kwargs)

    @Feat()
    def idn(self):
        return self.parse_query('I', format='I{:s}')

    @Action()
    def detect_axis(self):
        """ Find the number of axis available.

        The detection stops as soon as an empty controller is found.
        """
        pass


class MotionAxis(MotionAxisMultiple, BacklashMixing):
    def __del__(self):
        self.parent = None
        self.num = None

    def query(self, command, *, send_args=(None, None), recv_args=(None, None)):
        return self.parent.query('{:d}{}'.format(self.num, command),
                                 send_args=send_args, recv_args=recv_args)

    def write(self, command, *args, **kwargs):
        return self.parent.write('{:d}{}'.format(self.num, command),
                                 *args, **kwargs)

    @Feat()
    def idn(self):
        return self.query('ID?')

    @Action()
    def on(self):
        """Put axis on"""
        self.write('MO')

    @Action()
    def off(self):
        """Put axis on"""
        self.write('MF')

    @Feat(values={True: '1', False: '0'})
    def is_on(self):
        """
        :return: True is axis on, else false
        """
        return self.query('MO?')

    @Action(units='mm')
    def define_home(self, val=0):
        """Remap current position to home (0), or to new position

        :param val: new position"""
        self.write('DH%f' % val)

    @Action()
    def home(self):
        """Execute the HOME command"""
        self.write('OR')

    @Feat(units='mm')
    def position(self):
        return self.query('TP?')

    @position.setter
    def position(self, pos):
        """
        Waits until movement is done if self.wait_until_done = True.

        :param pos: new position
        """
        if not self.is_on:
            self.log_error('Axis not enabled. Not moving!')
            return

        # First do move to extra position if necessary
        self._set_position(pos, wait=self.wait_until_done)


    def __set_position(self, pos):
        """
        Move stage to a certain position
        :param pos: New position
        """
        self.write('PA%f' % (pos))
        self.last_set_position = pos

    @Feat(units='mm/s')
    def max_velocity(self):
        return float(self.query('VU?'))

    @max_velocity.setter
    def max_velocity(self, velocity):
        self.write('VU%f' % (velocity))

    @Feat(units='mm/s**2')
    def max_acceleration(self):
        return float(self.query('AU?'))

    @max_acceleration.setter
    def max_acceleration(self, velocity):
        self.write('AU%f' % (velocity))

    @Feat(units='mm/s')
    def velocity(self):
        return float(self.query('VA?'))

    @velocity.setter
    def velocity(self, velocity):
        """
        :param velocity: Set the velocity that the axis should use when moving
        :return:
        """
        self.write('VA%f' % (velocity))

    @Feat(units='mm/s**2')
    def acceleration(self):
        return float(self.query('VA?'))

    @acceleration.setter
    def acceleration(self, acceleration):
        """
        :param acceleration: Set the acceleration that the axis should use
                             when starting
        :return:
        """
        self.write('AC%f' % (acceleration))

    @Feat(units='mm/s')
    def actual_velocity(self):
        return float(self.query('TV'))

    @actual_velocity.setter
    def actual_velocity(self, val):
        raise NotImplementedError

    @Action()
    def stop(self):
        """Emergency stop"""
        self.write('ST')

    @Feat(values={True: '1', False: '0'})
    def motion_done(self):
        return self.query('MD?')

    # Not working yet, see https://github.com/hgrecco/lantz/issues/35
    # @Feat(values={Q_('encodercount'): 0,
    #                     Q_('motor step'): 1,
    #                     Q_('millimeter'): 2,
    #                     Q_('micrometer'): 3,
    #                     Q_('inches'): 4,
    #                     Q_('milli-inches'): 5,
    #                     Q_('micro-inches'): 6,
    #                     Q_('degree'): 7,
    #                     Q_('gradian'): 8,
    #                     Q_('radian'): 9,
    #                     Q_('milliradian'): 10,
    #                     Q_('microradian'): 11})
    @Feat()
    def units(self):
        ret = int(self.query(u'SN?'))
        return UNITS[ret]

    @units.setter
    def units(self, val):
        # No check implemented yet
        self.write('%SN%' % (self.num, UNITS.index(val)))
        super().units = val

    def _wait_until_done(self):
        # wait_time = convert_to('seconds', on_dimensionless='warn')(self.wait_time)
        time.sleep(self.wait_time)
        while not self.motion_done:
            time.sleep(self.wait_time)
        return True



# some stuff
"""

    lantz.drivers.agilent.e8257c
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Implementation of E8257C signal generator
    
    Author: Thomas Kiely
    
    Date: 6/16/2016
    
"""


import numpy as np
import lantz
from lantz import Action, Feat, DictFeat, ureg
from lantz.messagebased import MessageBasedDriver

class E8257C(MessageBasedDriver):
    
    DEFAULTS = {
        'COMMON': {
            'write_termination': '\n',
            'read_termination': '\n',
        }
    }
    
    @Feat(read_once=True)
    def idn(self):
        return self.query('*IDN?')

    @Feat(units='V')
    def lf_amplitude(self):
        """
        low frequency ampitude (BNC output)
        """
        return float(self.query('LFO:AMPL?'))
    
    @lf_amplitude.setter
    def lf_amplitude(self,value):
        self.write('LFO:AMPL {:.2f}'.format(value))
        
    @Feat()
    def rf_amplitude(self):
        """
        RF amplitude (Type N output)
        units are in dBm
        """
        return float(self.query('POW:AMPL?'))
    
    @rf_amplitude.setter
    def rf_amplitude(self,value):
        self.write('POW:AMPL {:.2f}'.format(value))
    
    @Feat(values={True: 1, False: 0})
    def lf_toggle(self):
        """
        enable or disable low frequency output
        """
        return float(self.query('LFO:STAT?'))

    @lf_toggle.setter
    def lf_toggle(self,value):
        self.write('LFO:STAT {:s}'.format(value))
    
    @Feat(values={True: '1', False: '0'})
    def rf_toggle(self):
        """
        enable or disable RF output
        """
        return self.query('OUTP:STAT?')
    
    @rf_toggle.setter
    def rf_toggle(self,value):
        self.write('OUTP:STAT {:s}'.format(value))
    
    @Feat(units='Hz')
    def rf_frequency(self):
        """
        signal frequency
        units are in hertz
        """
        return float(self.query('FREQ?'))
    
    @rf_frequency.setter
    def rf_frequency(self,value):
        self.write('FREQ {:.2f}'.format(value))
    
    @Feat(units='Hz')
    def lf_frequency(self):
        """
        signal frequency 0.5Hz-1MHz
        """
        return float(self.query('LFO:FUNC:FREQ?'))
    
    @lf_frequency.setter
    def lf_frequency(self,value):
        self.write('LFO:FUNC:FREQ {:.2f}'.format(value))
    
    #ALL EXCEPT MODEL UNR
    #@Feat(values={1,2})
    #def pll_loop_filter_mode(self):
    #    """
    #    sets PLL bandwidth to optimize phase noise
    #    1 optimizes below 10kHz, 2 optimizes above 10kHz
    #    """
    #    return float(self.query('FREQ:SYNT?'))
    #
    #@pll_loop_filter_mode.setter
    #def pll_loop_filter_mode(self,value):
    #    self.write('FREQ:SYNT {}'.format(value))
    
    @Feat()
    def rf_offset(self): #not entirely sure if this is rf
        """
        RF offset
        units are dB
        """
        return float(self.query('POW:OFFS?'))
    
    @rf_offset.setter
    def rf_offset(self,value):
        self.write('POW:OFFS {:.2f}'.format(value))
    
    @Feat(units='radians')
    def phase(self):
        """
        carrier phase
        from -Pi to Pi
        """
        return float(self.query('PHAS?'))
    
    @phase.setter
    def phase(self,value):
        self.write('PHAS {:.2f}'.format(value))
    
    @Action()
    def ref_phase(self):
        """
        sets current output phase as a zero reference
        """
        self.write('PHAS:REF')

    @Feat(values={True: 1, False: 0})
    def mod_toggle(self):
        """
        Modulation State
        """
        return float(self.query('OUTP:MOD?'))
        
    @mod_toggle.setter
    def mod_toggle(self, value):
        self.write('OUTP:MOD {}'.format(value))

#   ONLY WITH OPTION 002/602    
#    @Feat()
#    def mod_type(self):
#        """
#        Modulation State
#        """
#        return self.query('RAD:CUST:MOD:TYPE?')
#
#    @mod_type.setter
#    def mod_type(self, value):
#        self.write('RAD:CUST:MOD:TYPE {}'.format(value))


# -*- coding: utf-8 -*-
"""
    lantz.drivers.cobolt
    ~~~~~~~~~~~~~~~~~~~~

    :company: Cobolt.
    :description: DPSS lasers, diode laser modules, fiber pigtailed lasers.
    :website: http://www.cobolt.se/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .cobolt0601 import Cobolt0601

__all__ = ['Cobolt0601']



# Lantz driver for interfacing with Montana instruments cryostat.
# Includes testing code, which should work out of the box assuming you give
# it the correct IP address.
# Author: P. Mintun
# Date: 8/3/2016
# Version: 0.1

from lantz.driver import Driver
from lantz import Feat, DictFeat, Action

from lantz import Q_

import socket
import warnings

# class MontanaWarning(Warning):
#     """
#     """
#     def __init__(self):
#         super().__init__()

class Cryostation(Driver):

    def __init__(self, address, port=7773, timeout=2.0):
        super().__init__()
        self.address = address
        self.port = port
        self.timeout = timeout

    def initialize(self):
        """
        Initialize function for Cryostation communication port, uses Python
        sockets library to open socket communication with Cryostation.
        """
        #print('IP address:{}'.format(self.address))
        #print('Port:{}'.format(self.port))

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((self.address, self.port))
        self.socket.settimeout(self.timeout)

    def finalize(self):
        """
        Closes socket communication with Cryostation software.
        """
        self.socket.close()

    @Feat(values={True, False})
    def alarm_state(self):
        """
        Returns true or false, indicating the presence (T) or absence (F) of
        a system error.
        """
        error = self.send_and_recv('GAS')
        alarm = (error == 'T')
        return alarm

    @Feat(units='millitorr')
    def chamber_pressure(self):
        """
        Returns the chamber pressure in mTorr, or -0.1 if the pressure is
        unavailable.
        """
        return float(self.send_and_recv('GCP', raise_warning=True))

    @Feat(units='kelvin')
    def platform_temperature(self):
        """
        Returns the current platform temperature in K, or -0.100 if the current
        temperature is unavailable.
        """
        return float(self.send_and_recv('GPT', raise_warning=True))

    @Feat(units='kelvin')
    def platform_stability(self):
        """
        Returns the platform stability in K, or -0.100 if unavailable.
        """
        return float(self.send_and_recv('GPS'))

    @Feat(units='watts')
    def platform_heater_pow(self):
        """
        Returns the current platform heater power in W, or -0.100 if
        unavailable.
        """
        return float(self.send_and_recv('GPHP', raise_warning=True))

    @Feat(units='kelvin')
    def stage_1_temperature(self):
        """
        Returns the current stage 1 temperature in K, or -0.100 if the current
        temperature is unavailable.
        """
        return float(self.send_and_recv('GS1T', raise_warning=True))

    @Feat(units='watts')
    def stage_1_heater_pow(self):
        """
        Returns the current stage 1 heater power in W, or -0.100 if
        unavailable.
        """
        return float(self.send_and_recv('GS1HP', raise_warning=True))

    @Feat(units='kelvin')
    def stage_2_temperature(self):
        """
        Returns the current stage 2 temperature in K, or -0.100 if the current
        temperature is unavailable.
        """
        return float(self.send_and_recv('GS2T', raise_warning=True))

    @Feat(units='kelvin')
    def sample_stability(self):
        """
        Returns the sample stability in K, or -0.100 if unavailable.
        """
        return float(self.send_and_recv('GSS'))

    @Feat(units='kelvin')
    def sample_temperature(self):
        """
        Returns the sample temperature in K, or -0.100 if unavailable.
        """
        return float(self.send_and_recv('GST', raise_warning=True))

    @Feat(units='kelvin')
    def temp_setpoint(self):
        """
        Returns the temperature setpoint of the Cryostation software.
        """
        return float(self.send_and_recv('GTSP', raise_warning=True))

    @temp_setpoint.setter
    def temp_setpoint(self, setpoint_kelvin):
        """
        Sets the temperature setpoint of the Cryostation software.
        """
        return self.send_and_recv('STSP{0:.2f}'.format(setpoint_kelvin))

    @Feat(units='kelvin')
    def user_temperature(self):
        """
        Returns the user thermometer temperature in K, or -0.100 if unavailable.
        """
        return float(self.send_and_recv('GUT'))

    @Feat(units='kelvin')
    def user_stability(self):
        """
        Returns the user thermometer stability in K, or -0.100 if unavailable.
        """
        return float(self.send_and_recv('GUS'))

    @Action()
    def cool_down(self):
        """
        Initializes cooldown to temperature setpoint.
        """
        return self.send_and_recv('SCD')

    @Action()
    def standby(self):
        """
        Puts the system into standby mode (see manual for details).
        """
        return self.send_and_recv('SSB')

    @Action()
    def stop(self):
        """
        Returns the status of the last command.
        If OK, the system cannot stop at this time.
        """
        return self.send_and_recv('STP')

    @Action()
    def warm_up(self):
        """
        Starts the system warmup to room temperature.
        """
        return self.send_and_recv('SWU')

    def send_and_recv(self, message, raise_warning=False):
        """
        Params:
            message = command to be sent to Cryostation
            return_bytes = number of bytes to be stripped off return string
        """
        buffer_size = 1024
        m1 = '0{}{}'.format(str(len(message)), message)

        self.socket.send(m1.encode())
        data = str()
        received = self.socket.recv(buffer_size)
        if received:
            data = received.decode()

        if ((data[2:] == '-0.100') and raise_warning):

            warnings.warn('Unable to return parameter from command {}.'.format(message))
            return 0

        return data[2:]



# -*- coding: utf-8 -*-
"""
    lantz.drivers.ni.daqmx.base
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of base classes for Channels, Tasks and Devices


    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz import Feat, Action
from lantz import errors
from lantz.foreign import LibraryDriver, RetValue, RetStr

from .constants import Constants, Types

default_buf_size = 2048

_SAMPLE_MODES = {'finite': Constants.Val_FiniteSamps,
                 'continuous': Constants.Val_ContSamps,
                 'hwtimed': Constants.Val_HWTimedSinglePoint}

_BUS_TYPES = {'PCI': Constants.Val_PCI, 'PCIe': Constants.Val_PCIe,
              'PXI': Constants.Val_PXI, 'SCXI': Constants.Val_SCXI,
              'PCCard': Constants.Val_PCCard, 'USB': Constants.Val_USB,
              'UNKNOWN': Constants.Val_Unknown}

_SIGNAL_TYPES = {'sample_clock': Constants.Val_SampleClock,
                 'sample_complete': Constants.Val_SampleCompleteEvent,
                 'change_direction': Constants.Val_ChangeDetectionEvent,
                 'counter_output': Constants.Val_CounterOutputEvent}

_EDGE_TYPES = {'rising': Constants.Val_Rising, 'falling': Constants.Val_Falling}

_SLOPE_TYPES = {'rising': Constants.Val_RisingSlope, 'falling': Constants.Val_FallingSlope}

_WHEN_WINDOW = {'entering': Constants.Val_EnteringWin, 'leaving': Constants.Val_LeavingWin}

_WHEN_TRIGGER_DIG = {'high': Constants.Val_High, 'low': Constants.Val_Low}

_WHEN_TRIGGER_ALVL = {'above': Constants.Val_AboveLvl, 'below': Constants.Val_BelowLvl}

_WHEN_TRIGGER_AWIN = {'inside': Constants.Val_InsideWin, 'outside': Constants.Val_OutsideWin}

_WHEN_MATCH = {True: Constants.Val_PatternMatches, False: Constants.Val_PatternDoesNotMatch}

_TRIGGER_TYPES = {'digital_level': Constants.Val_DigLvl,
                  'analog_level': Constants.Val_AnlgLvl,
                  'analog_window': Constants.Val_AnlgWin}

_CHANNEL_TYPES = {'AI': Constants.Val_AI, 'AO': Constants.Val_AO,
                  'DI': Constants.Val_DI, 'DO': Constants.Val_DO,
                  'CI': Constants.Val_CI, 'CO': Constants.Val_CO}

class _Base(LibraryDriver):
    """Base class for NIDAQmx
    """

    LIBRARY_NAME = 'nicaiu', 'nidaqmx'
    LIBRARY_PREFIX = 'DAQmx'

    _DEVICES = {}
    _TASKS = {}
    _CHANNELS = {}

    def _get_error_string(self, error_code):
        size = self.lib.GetErrorString(error_code, None, 0)
        if size <= 0:
            raise errors.InstrumentError('Could not retrieve error string.')
        err, msg = self.lib.GetErrorString(error_code, *RetStr(size))
        if err < 0:
            raise errors.InstrumentError('Could not retrieve error string.')
        return msg

    def _get_error_extended_error_info(self):
        size = self.lib.GetExtendedErrorInfo(None, 0)
        if size <= 0:
            raise errors.InstrumentError('Could not retrieve extended error info.')
        err, msg = self.lib.GetExtendedErrorInfo(*RetStr(size))
        if err < 0:
            raise errors.InstrumentError('Could not retrieve extended error info.')
        return msg

    def _return_handler(self, func_name, ret_value):
        if ret_value < 0 and func_name not in ('GetErrorString', 'GetExtendedErrorInfo'):
            msg = self._get_error_string(ret_value)
            raise errors.InstrumentError(msg)
        return ret_value

    def __get_fun(self, name):
        return getattr(self.lib, name.format(self.operation_direction.title()))

    def _add_types(self):

        super()._add_types()
        T = Types
        self.lib.CreateAIVoltageChan.argtypes = [T.TaskHandle, T.string, T.string, T.int32, T.float64, T.float64, T.int32, T.string]
        self.lib.ReadAnalogScalarF64.argtypes = [T.TaskHandle, T.float64, T._, T._]

class _ObjectDict(object):

    def __init__(self, key_fun, obj_creator, dictionary=None):
        self.key_fun = key_fun
        self.obj_creator = obj_creator
        self._internal = {} if dictionary is None else dictionary

    def __getitem__(self, item):

        if item in self._internal:
            return self._internal[item]

        if item not in self:
            raise KeyError('{} not found'.format(item))

        value = self.obj_creator(item)

        self._internal[item] = value

        return value

    def __len__(self):
        return sum((1 for key in self.keys()))

    def __contains__(self, item):
        return item in self.key_fun()

    def keys(self):
        for key in self.key_fun():
            yield key

    def values(self):
        for key in self.keys():
            yield self[key]

    def items(self):
        for key in self.keys():
            yield key, self[key]


class System(_Base):
    """NI-DAQmx System
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.devices = _ObjectDict(self._device_names, Device, self._DEVICES)
        self.tasks = _ObjectDict(self._task_names, Task, self._TASKS)
        self.channels = _ObjectDict(self._channel_names, Channel, self._CHANNELS)

    @Feat(read_once=True)
    def version(self):
        """Version of installed NI-DAQ library. 
        """
        err, major  = self.lib.GetSysNIDAQMajorVersion(RetValue('u32'))
        err, minor  = self.lib.GetSysNIDAQMinorVersion(RetValue('u32'))
        return major, minor

    def _device_names(self):
        """Return a tuple containing the names of all global devices installed in the system.
        """
        err, buf = self.lib.GetSysDevNames(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    def _task_names(self):
        """Return a tuple containing the names of all global tasks saved in the system.
        """
        err, buf = self.lib.GetSysTasks(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    def _channel_names(self):
        """Return a tuple containing the names of all global channels saved in the system.
        """
        err, buf = self.lib.GetSysGlobalChans(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names


class Device(_Base):
    """Device
    """

    def __init__(self, device_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.device_name = device_name

    def _preprocess_args(self, name, *args):
        """Injects device_name to all call to the library
        """
        if name in ('GetErrorString', 'GetExtendedErrorInfo'):
            return super()._preprocess_args(name, *args)
        else:
            return super()._preprocess_args(name, *((self.device_name, ) + args))

    @Feat(read_once=True)
    def is_simulated(self):
        """Return True if it is a simulated device
        """
        err, value = self.lib.GetDevIsSimulated(RetValue('u32'))
        return value != 0

    @Feat(read_once=True)
    def product_type(self):
        """Return the product name of the device.
        """
        err, buf = self.lib.GetDevProductType(*RetStr(default_buf_size))
        return buf

    @Feat(read_once=True)
    def product_number(self):
        """Return the unique hardware identification number for the device.
        """
        err, value = self.lib.GetDevProductNum(RetValue('u32'))
        return value

    @Feat(read_once=True)
    def serial_number (self):
        """Return the serial number of the device. This value is zero
        if the device does not have a serial number.
        """
        err, value = self.lib.GetDevSerialNum(RetValue('u32'))
        return value

    @Feat(read_once=True)
    def analog_input_channels(self):
        """Return a tuple with the names of the analog input
        physical channels available on the device.
        """

        err, buf = self.lib.GetDevAIPhysicalChans(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    @Feat(read_once=True)
    def analog_output_channels(self):
        """Return a tuple with the names of the analog output
        physical channels available on the device.
        """
        err, buf = self.lib.GetDevAOPhysicalChans(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    @Feat(read_once=True)
    def digital_input_lines(self):
        """Return a tuple with the names of the digital input lines
        physical channels available on the device.
        """
        err, buf = self.lib.GetDevDILines(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    @Feat(read_once=True)
    def digital_output_lines(self):
        """Return a tuple with the names of the digital lines
        ports available on the device.
        """
        err, buf = self.lib.GetDevDOLines(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    @Feat(read_once=True)
    def digital_input_ports(self):
        """Return a tuple with the names of the digital input
        ports available on the device.
        """
        err, buf = self.lib.GetDevDIPorts(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    @Feat(read_once=True)
    def digital_output_ports(self):
        """Return a tuple with the names of the digital output
        ports available on the device.
        """
        err, buf = self.lib.GetDevDOPorts(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    @Feat(read_once=True)
    def counter_input_channels(self):
        """Return a tuple with the names of the counter input
        physical channels available on the device.
        """
        err, buf = self.lib.GetDevCIPhysicalChans(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    @Feat(read_once=True)
    def counter_output_channels(self):
        """Return a tuple with the names of the counter input
        physical channels available on the device.
        """
        err, buf = self.lib.GetDevCOPhysicalChans(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    @Feat(read_once=True, values=_BUS_TYPES)
    def bus_type(self):
        """Return the bus type of the device.
        """
        err, value = self.lib.GetDevBusType(RetValue('i32'))
        return value

    @Feat(read_once=True)
    def pci_bus_number (self):
        """Return the PCI bus number of the device.
        """
        err, value = self.lib.GetDevPCIBusNum(RetValue('i32'))
        return value

    @Feat(read_once=True)
    def pci_device_number (self):
        """Return the PCI slot number of the device.
        """
        err, value = self.lib.GetDevPCIDevNum(RetValue('i32'))
        return value

    @Feat(read_once=True)
    def pxi_slot_number(self):
        """Return the PXI slot number of the device.
        """
        err, value = self.lib.GetDevPXISlotNum(RetValue('u32'))
        return value

    @Feat(read_once=True)
    def pxi_chassis_number(self):
        """Return the PXI chassis number of the device, as identified
        in MAX.
        """
        err, value = self.lib.GetDevPXIChassisNum(RetValue('u32'))
        return value

    @Feat(read_once=True)
    def bus_info(self):
        t = self.bus_type
        if t in ('PCI', 'PCIe'):
            return '%s (bus=%s, device=%s)' % (t, self.pci_bus_number, self.pci_device_number)
        if t == 'PXI':
            return '%s (chassis=%s, slot=%s)' % (t, self.pxi_chassis_number, self.pxi_slot_number)
        return t

    @Action()
    def reset(self):
        """Stops and deletes all tasks on a device and rests outputs to their defaults
        """
        return self.lib.ResetDevice()


class Task(_Base):
    """A task is a collection of one or more virtual channels with timing,
    triggering, and other properties. Conceptually, a task represents a
    measurement or generation you want to perform. All channels in a task
    must be of the same I/O type, such as analog input or counter output.

    However, a task can include channels of different measurement types,
    such as an analog input temperature channel and an analog input voltage
    channel. For most devices, only one task per subsystem can run at once,
    but some devices can run multiple tasks simultaneously. With some devices,
    you can include channels from multiple devices in a task.

    :param name: Name assigned to the task (This can be changed by the
                 library. The final name will be stored in name attribute)

    """

    _REGISTRY = {}

    @classmethod
    def register_class(cls, klass):
        cls._REGISTRY[klass.IO_TYPE] = klass

    @classmethod
    def typed_task(cls, io_type):
        return cls._REGISTRY[io_type]

    def _load_task(self, name):
        err, self.__task_handle = self.lib.LoadTask(name, RetValue('u32'))
        self.name = name
        self.log_debug('Loaded task with {} ({})'.format(self.name, self.__task_handle))

    def _create_task(self, name):
        err, self.__task_handle = self.lib.CreateTask(name, RetValue('u32'))
        err, self.name = self.lib.GetTaskName(*RetStr(default_buf_size))
        self.log_debug('Created task with {} ({})'.format(self.name, self.__task_handle))

    def __init__(self, name='', *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        if not name:
            self._create_task(name)
        else:
            try:
                self._load_task(name)
            except Exception as e:
                self._create_task(name)

        self.sample_mode = None
        self.channels = _ObjectDict(self._channel_names, self._create_channel_from_name, self._CHANNELS)
        self.devices = _ObjectDict(self._device_names, Device, self._DEVICES)

    @property
    def task_handle(self):
        return self.__task_handle

    def _preprocess_args(self, name, *args):
        """Injects device_name to all call to the library
        """
        if name in ('GetErrorString', 'GetExtendedErrorInfo', 'LoadTask', 'CreateTask'):
            return super()._preprocess_args(name, *args)
        else:
            return super()._preprocess_args(name, *((self.task_handle, ) + args))

    def _create_channel_from_name(self, name):
        return Channel(self, name=name)

    def _channel_names(self):
        """Return a tuple with the names of all virtual channels in the task.
        """
        err, buf = self.lib.GetTaskChannels(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    def _device_names(self):
        """Return a tuple with the names of all devices in the task.
        """
        err, buf = self.lib.GetTaskDevices(*RetStr(default_buf_size))
        names = tuple(n.strip() for n in buf.split(',') if n.strip())
        return names

    @Feat()
    def io_type(self):
        for name, channel in self.channels.items():
            return channel.io_type
        else:
            return None

    def add_channel(self, channel):
        if not isinstance(channel, Channel):
            raise TypeError('Only channels may be added to a task.')

        if channel.task is self.channels:
            return
        elif channel.task is None:
            channel.task = self
        else:
            raise ValueError('Cannot add a channel that is already in another task')

    def execute_fun(self, func_name, *args):
        return getattr(self.lib, func_name)(*args)

    def clear(self):
        """Clear the task.

        Before clearing, this function stops the task, if necessary,
        and releases any resources reserved by the task. You cannot
        use a task once you clear the task without recreating or
        reloading the task.

        If you use the DAQmxCreateTask function or any of the NI-DAQmx
        Create Channel functions within a loop, use this function
        within the loop after you finish with the task to avoid
        allocating unnecessary memory.
        """
        if self.task_handle:
            self.lib.ClearTask()
            self.__task_handle = None

    __del__ = clear

    @Feat()
    def is_done(self):
        """Queries the status of the task and indicates if it completed
        execution. Use this function to ensure that the specified
        operation is complete before you stop the task.
        """
        err, value = self.lib.IsTaskDone(RetValue('u32'))
        return value != 0

    # States

    @Action()
    def start(self):
        """Start the task

        Transitions the task from the committed state to the running
        state, which begins measurement or generation. Using this
        function is required for some applications and optional for
        others.

        If you do not use this function, a measurement task starts
        automatically when a read operation begins. The autoStart
        parameter of the NI-DAQmx Write functions determines if a
        generation task starts automatically when you use an NI-DAQmx
        Write function.

        If you do not call StartTask and StopTask when you
        call NI-DAQmx Read functions or NI-DAQmx Write functions
        multiple times, such as in a loop, the task starts and stops
        repeatedly. Starting and stopping a task repeatedly reduces
        the performance of the application.
        """

        self.lib.StartTask()

    @Action()
    def stop(self):
        """Stop the task.

        Stop the task and returns it to the state it was in before
        you called StartTask or called an NI-DAQmx Write function with
        autoStart set to TRUE.

        If you do not call StartTask and StopTask when you call
        NI-DAQmx Read functions or NI-DAQmx Write functions multiple
        times, such as in a loop, the task starts and stops
        repeatedly. Starting and stopping a task repeatedly reduces
        the performance of the application.

        """
        self.lib.StopTask()

    @Action()
    def verify(self):
        """Verifies that all task parameters are valid for the hardware.
        """
        self.alter_state('verify')

    @Action()
    def commit(self):
        """Programs the hardware as much as possible according
        to the task configuration.
        """
        self.alter_state('commit')

    @Action()
    def reserve(self):
        """Reserves the hardware resources needed for the
        task. No other tasks can reserve these same resources.
        """
        self.alter_state('reserve')

    @Action()
    def unreserve(self):
        """Release all reserved resources.
        """
        self.alter_state('unreserve')

    @Action()
    def abort(self):
        """Abort an operation, such as Read or Write, that is currently active.

        The task is put into an unstable but recoverable state. To recover the
        task, call Start to restart the task or call Stop to reset the task
        without starting it.
        """
        self.alter_state('abort')

    @Action(values={'start': Constants.Val_Task_Start, 'stop': Constants.Val_Task_Stop,
                    'verify': Constants.Val_Task_Verify, 'commit': Constants.Val_Task_Commit,
                    'reserve': Constants.Val_Task_Reserve, 'unreserve': Constants.Val_Task_Unreserve,
                    'abort': Constants.Val_Task_Abort})
    def alter_state(self, new_state):
        """Alters the state of a task according to the action you
        specify. To minimize the time required to start a task, for
        example, DAQmxTaskControl can commit the task prior to
        starting.

        :param new_state:
        """
        self.lib.TaskControl(new_state)

    _register_every_n_samples_event_cache = None

    def register_every_n_samples_event(self, func, samples=1, options=0, cb_data=None):
        """Register a callback function to receive an event when the
        specified number of samples is written from the device to the
        buffer or from the buffer to the device. This function only
        works with devices that support buffered tasks.

        When you stop a task explicitly any pending events are
        discarded. For example, if you call DAQmxStopTask then you do
        not receive any pending events.

        :param func: The function that you want DAQmx to call when the event
          occurs. The function you pass in this parameter must have
          the following prototype::

            def func(task, event_type, samples, cb_data):
                ...
                return 0

          Upon entry to the callback, the task parameter contains the
          handle to the task on which the event occurred. The
          event_type parameter contains the value you passed in the
          event_type parameter of this function. The samples parameter
          contains the value you passed in the samples parameter of
          this function. The cb_data parameter contains the value you
          passed in the cb_data parameter of this function.

        :param samples: The number of samples after which each event should occur.

        :param options:

        :param cb_data:

        See also: register_signal_event, register_done_event
        """

        if self.operation_direction == 'input':
            event_type = Constants.Val_Acquired_Into_Buffer
        else:
            event_type = Constants.Val_Transferred_From_Buffer

        if options == 'sync':
            options = Constants.Val_SynchronousEventCallbacks

        if func is None:
            c_func = None # to unregister func
        else:
            if self._register_every_n_samples_event_cache is not None:
                # unregister:
                self.register_every_n_samples_event(None, samples=samples, options=options, cb_data=cb_data)
                # TODO: check the validity of func signature
            # TODO: use wrapper function that converts cb_data argument to given Python object
            c_func = EveryNSamplesEventCallback_map[self.CHANNEL_TYPE](func)

        self._register_every_n_samples_event_cache = c_func

        self.lib.RegisterEveryNSamplesEvent(event_type, uInt32(samples), uInt32(options), c_func, cb_data)

    _register_done_event_cache = None

    def register_done_event(self, func, options=0, cb_data=None):
        """Register a callback function to receive an event when a task
        stops due to an error or when a finite acquisition task or
        finite generation task completes execution. A Done event does
        not occur when a task is stopped explicitly, such as by
        calling DAQmxStopTask.


        :param func:

          The function that you want DAQmx to call when the event
          occurs.  The function you pass in this parameter must have
          the following prototype::

            def func(task, status, cb_data = None):
                ...
                return 0

          Upon entry to the callback, the taskHandle parameter
          contains the handle to the task on which the event
          occurred. The status parameter contains the status of the
          task when the event occurred. If the status value is
          negative, it indicates an error. If the status value is
          zero, it indicates no error. If the status value is
          positive, it indicates a warning. The callbackData parameter
          contains the value you passed in the callbackData parameter
          of this function.

        :param options : {int, 'sync'}

          Use this parameter to set certain options. You can
          combine flags with the bitwise-OR operator ('|') to set
          multiple options. Pass a value of zero if no options need to
          be set.

          'sync' - The callback function is called in the thread which
          registered the event. In order for the callback to occur,
          you must be processing messages. If you do not set this
          flag, the callback function is called in a DAQmx thread by
          default.

          Note: If you are receiving synchronous events faster than
          you are processing them, then the user interface of your
          application might become unresponsive.

        :param cb_data:

          A value that you want DAQmx to pass to the callback function
          as the function data parameter. Do not pass the address of a
          local variable or any other variable that might not be valid
          when the function is executed.

        See also

        register_signal_event, register_every_n_samples_event
        """
        if options=='sync':
            options = Constants.Val_SynchronousEventCallbacks

        if func is None:
            c_func = None
        else:
            if self._register_done_event_cache is not None:
                self.register_done_event(None, options=options, cb_data=cb_data)
                # TODO: check the validity of func signature
            c_func = DoneEventCallback_map[self.CHANNEL_TYPE](func)
        self._register_done_event_cache = c_func

        self.lib.RegisterDoneEvent(uInt32(options), c_func, cb_data)

    def operation_direction(self):
        return 'input' if self.CHANNEL_TYPE else 'output'

    _register_signal_event_cache = None

    @Action(values=(None, _SIGNAL_TYPES, None, None))
    def register_signal_event(self, func, signal, options=0, cb_data=None):
        """Registers a callback function to receive an event when the
        specified hardware event occurs.

        When you stop a task explicitly any pending events are
        discarded. For example, if you call DAQmxStopTask then you do
        not receive any pending events.

        :param func:

          The function that you want DAQmx to call when the event
          occurs. The function you pass in this parameter must have the
          following prototype::

            def func(task, signalID, cb_data):
              ...
              return 0

          Upon entry to the callback, the task parameter contains the
          handle to the task on which the event occurred. The signalID
          parameter contains the value you passed in the signal
          parameter of this function. The cb_data parameter contains
          the value you passed in the cb_data parameter of this
          function.

        :param signal: {'sample_clock', 'sample_complete', 'change_detection', 'counter_output'}

          The signal for which you want to receive results:

          'sample_clock' - Sample clock
          'sample_complete' - Sample complete event
          'change_detection' - Change detection event
          'counter_output' - Counter output event

        :param options:

        :param cb_data:

        See also: register_done_event, register_every_n_samples_event
        """

        if options == 'sync':
            options = Constants.Val_SynchronousEventCallbacks

        if func is None:
            c_func = None
        else:
            if self._register_signal_event_cache is not None:
                self._register_signal_event(None, signal=signal, options=options, cb_data=cb_data)
                # TODO: check the validity of func signature
            c_func = SignalEventCallback_map[self.CHANNEL_TYPE](func)
        self._register_signal_event_cache = c_func
        self.lib.RegisterSignalEvent(signal, uInt32(options), c_func, cb_data)


    @Action(values=(str, str, _SAMPLE_MODES, None))
    def configure_timing_change_detection(self, rising_edge_channel='', falling_edge_channel='',
                                          sample_mode='continuous', samples_per_channel=1000):
        """Configures the task to acquire samples on the rising and/or
        falling edges of the lines or ports you specify.
        """

        self.lib.CfgChangeDetectionTiming(rising_edge_channel, falling_edge_channel,
                                          sample_mode, samples_per_channel)


    @Action(values=(_SAMPLE_MODES, None))
    def configure_timing_handshaking(self, sample_mode='continuous', samples_per_channel=1000):
        """Determines the number of digital samples to acquire or
        generate using digital handshaking between the device and a
        peripheral device.
        """
        self.samples_per_channel = samples_per_channel
        self.sample_mode = sample_mode
        self.lib.CfgHandshakingTiming(sample_mode, samples_per_channel)

    @Action(values=(_SAMPLE_MODES, None))
    def configure_timing_implicit(self, sample_mode='continuous', samples_per_channel=1000):
        """Sets only the number of samples to acquire or generate without
        specifying timing. Typically, you should use this function
        when the task does not require sample timing, such as tasks
        that use counters for buffered frequency measurement, buffered
        period measurement, or pulse train generation.
        """
        self.samples_per_channel = samples_per_channel
        self.sample_mode = sample_mode
        self.lib.CfgImplicitTiming(self, sample_mode, samples_per_channel)

    @Action(values=(str, None, _EDGE_TYPES, _SAMPLE_MODES, None))
    def configure_timing_sample_clock(self, source='on_board_clock', rate=1, active_edge='rising',
                                      sample_mode='continuous', samples_per_channel=1000):
        """Set the source of the Sample Clock, the rate of the Sample
        Clock, and the number of samples to acquire or generate.

        :param source:

            The source terminal of the Sample Clock. To use the
            internal clock of the device, use None or use
            'OnboardClock'.

        :param rate:

            The sampling rate in samples per second. If you use an
            external source for the Sample Clock, set this value to
            the maximum expected rate of that clock.

        :param active_edge:

            Specifies on which edge of the clock to
            acquire or generate samples:

              'rising' - Acquire or generate samples on the rising edges
              of the Sample Clock.

              'falling' - Acquire or generate samples on the falling
              edges of the Sample Clock.

        :param sample_mode: {'finite', 'continuous', 'hwtimed'}

            Specifies whether the task acquires or
            generates samples continuously or if it acquires or
            generates a finite number of samples:

              'finite' - Acquire or generate a finite number of samples.

              'continuous' - Acquire or generate samples until you stop the task.

              'hwtimed' - Acquire or generate samples continuously
              using hardware timing without a buffer. Hardware timed
              single point sample mode is supported only for the
              sample clock and change detection timing types.

        :param samples_per_channel:

            The number of samples to acquire or generate for each
            channel in the task if `sample_mode` is 'finite'.  If
            sample_mode is 'continuous', NI-DAQmx uses this value to
            determine the buffer size.
        """
        if source == 'on_board_clock':
            source = None
        self.samples_per_channel = samples_per_channel
        self.sample_mode = sample_mode
        self.lib.CfgSampClkTiming(source, float64(rate), active_edge, sample_mode, samples_per_channel)

    def configure_timing_burst_handshaking_export_clock(self, *args, **kws):
        """
        Configures when the DAQ device transfers data to a peripheral
        device, using the DAQ device's onboard sample clock to control
        burst handshaking timing.
        """
        raise NotImplementedError

    def configure_timing_burst_handshaking_import_clock(self, *args, **kws):
        """
        Configures when the DAQ device transfers data to a peripheral
        device, using an imported sample clock to control burst
        handshaking timing.
        """
        raise NotImplementedError

    @Action(values=(str, _SLOPE_TYPES, None))
    def configure_trigger_analog_edge_start(self, source, slope='rising', level=1.0):
        """
        Configures the task to start acquiring or generating samples
        when an analog signal crosses the level you specify.

        :param source:

          The name of a channel or terminal where there is an analog
          signal to use as the source of the trigger. For E Series
          devices, if you use a channel name, the channel must be the
          first channel in the task. The only terminal you can use for
          E Series devices is PFI0.

        :param slope:

          Specifies on which slope of the signal to start acquiring or
          generating samples when the signal crosses trigger level:

            'rising' - Trigger on the rising slope of the signal.

            'falling' - Trigger on the falling slope of the signal.

        :param level:

          The threshold at which to start acquiring or generating
          samples. Specify this value in the units of the measurement
          or generation. Use trigger slope to specify on which slope
          to trigger at this threshold.
        """

        self.lib.CfgAnlgEdgeStartTrig(source, slope, level)

    @Action(values=(str, _WHEN_WINDOW, None, None))
    def configure_trigger_analog_window_start(self, source, when='entering', top=1.0, bottom=-1.0):
        """Configure the task to start acquiring or generating samples
        when an analog signal enters or leaves a range you specify.

        :param source:

          The name of a virtual channel or terminal where there
          is an analog signal to use as the source of the trigger.

          For E Series devices, if you use a virtual channel, it must
          be the first channel in the task. The only terminal you can
          use for E Series devices is PFI0.

        :param when : {'entering', 'leaving'}

          Specifies whether the task starts measuring or generating
          samples when the signal enters the window or when it leaves
          the window. Use `bottom` and `top` to specify the limits of
          the window.

        :param top : float

          The upper limit of the window. Specify this value in the
          units of the measurement or generation.

        :param bottom : float

          The lower limit of the window. Specify this value in the
          units of the measurement or generation.
        """
        self.lib.CfgAnlgWindowStartTrig(self, source, when, top, bottom)

    @Action(values=(str, _EDGE_TYPES))
    def configure_trigger_digital_edge_start(self, source, edge='rising'):
        """Configure the task to start acquiring or generating samples
        on a rising or falling edge of a digital signal.

        :param source:

          The name of a terminal where there is a digital signal to
          use as the source of the trigger.

        :param edge: {'rising', 'falling'}

          Specifies on which edge of a digital signal to start
          acquiring or generating samples: rising or falling edge(s).
        """

        self.lib.CfgDigEdgeStartTrig(self, source, edge)

    @Action(values=(str, str, _WHEN_MATCH))
    def configure_trigger_digital_pattern_start(self, source, pattern, match=True):
        """Configure a task to start acquiring or generating samples
        when a digital pattern is matched.

        :param source:

          Specifies the physical channels to use for pattern
          matching. The order of the physical channels determines the
          order of the pattern. If a port is included, the order of
          the physical channels within the port is in ascending order.

        :param pattern:

          Specifies the digital pattern that must be met for the
          trigger to occur.

        :param when: {'matches', 'does_not_match'}

          Specifies the conditions under which the trigger
          occurs: pattern matches or not.
        """
        self.lib.CfgDigPatternStartTrig(self, source, pattern, match)

    def configure_trigger_disable_start(self):
        """
        Configures the task to start acquiring or generating samples
        immediately upon starting the task.

        Returns
        -------

          success_status : bool
        """
        self.lib.DisableStartTrig()

    @Action(values=(str, _SLOPE_TYPES, None, None))
    def configure_analog_edge_reference_trigger(self, source, slope='rising',level=1.0, pre_trigger_samps=0):
        """Configure the task to stop the acquisition when the device
        acquires all pretrigger samples, an analog signal reaches the
        level you specify, and the device acquires all post-trigger samples.

        :param source:

          The name of a channel or terminal where there is an analog
          signal to use as the source of the trigger. For E Series
          devices, if you use a channel name, the channel must be the
          first channel in the task. The only terminal you can use for
          E Series devices is PFI0.

        :param slope:

          Specifies on which slope of the signal to start acquiring or
          generating samples when the signal crosses trigger level:

            'rising' - Trigger on the rising slope of the signal.

            'falling' - Trigger on the falling slope of the signal.

        :param level:
          The threshold at which to start acquiring or generating
          samples. Specify this value in the units of the measurement
          or generation. Use trigger slope to specify on which slope
          to trigger at this threshold.

        :param pre_trigger_samps:

          The minimum number of samples per channel to acquire before
          recognizing the Reference Trigger. The number of posttrigger
          samples per channel is equal to number of samples per channel
          in the NI-DAQmx Timing functions minus pretriggerSamples.
        """

        self.lib.CfgAnlgEdgeRefTrig(source, slope, level, pre_trigger_samps)


    @Feat(values=(str, _WHEN_WINDOW, None, None, None))
    def configure_analog_window_reference_trigger(self, source, when='entering',top=1.0, bottom=1.0, pre_trigger_samps=0):
        """Configure the task to stop the acquisition when the device
        acquires all pretrigger samples, an analog signal enters or
        leaves a range you specify, and the device acquires all
        post-trigger samples.


        :param source : str

          The name of a channel or terminal where there is an analog
          signal to use as the source of the trigger. For E Series
          devices, if you use a channel name, the channel must be the
          first channel in the task. The only terminal you can use for
          E Series devices is PFI0.

        :param when : {'entering', 'leaving'}

          Specifies whether the Reference Trigger occurs when the signal
          enters the window or when it leaves the window. Use
          bottom and top to specify the limits of the window.

            'entering' - Trigger when the signal enters the window.

            'leaving' - Trigger when the signal leaves the window.

        :param top : float

          The upper limit of the window. Specify this value in the
          units of the measurement or generation.

        :param bottom : float

          The lower limit of the window. Specify this value in the
          units of the measurement or generation.

        :param pre_trigger_samps : uint32

          The minimum number of samples per channel to acquire before
          recognizing the Reference Trigger. The number of posttrigger
          samples per channel is equal to number of samples per channel
          in the NI-DAQmx Timing functions minus pretriggerSamples.
        """

        self.lib.CfgAnlgWindowRefTrig(source, when, top, bottom, pre_trigger_samps)

    @Action(values=(str, _SLOPE_TYPES, None))
    def configure_digital_edge_reference_trigger(self, source, slope='rising', pre_trigger_samps=0):
        """Configures the task to stop the acquisition when the device
        acquires all pretrigger samples, detects a rising or falling
        edge of a digital signal, and acquires all posttrigger samples.

        :param source:

          The name of a channel or terminal where there is an analog
          signal to use as the source of the trigger. For E Series
          devices, if you use a channel name, the channel must be the
          first channel in the task. The only terminal you can use for
          E Series devices is PFI0.

        :param slope:

          Specifies on which slope of the signal to start acquiring or
          generating samples when the signal crosses trigger level:

            'rising' - Trigger on the rising slope of the signal.

            'falling' - Trigger on the falling slope of the signal.

        :param pre_trigger_samps:

          The minimum number of samples per channel to acquire before
          recognizing the Reference Trigger. The number of posttrigger
          samples per channel is equal to number of samples per channel
          in the NI-DAQmx Timing functions minus pretriggerSamples.
        """
        if not source.startswith('/'): # source needs to start with a '/' TODO WHY?
            source = '/'  + source
        self.lib.CfgDigEdgeRefTrig(source, slope, pre_trigger_samps)


    @Action(values=(str, str, _WHEN_MATCH, None))
    def configure_digital_pattern_reference_trigger(self, source, pattern, match=True, pre_trigger_samps=0):
        """Configure the task to stop the acquisition when the device
        acquires all pretrigger samples, matches or does not match
        a digital pattern, and acquires all posttrigger samples.

        :param source:

          The name of a channel or terminal where there is an analog
          signal to use as the source of the trigger. For E Series
          devices, if you use a channel name, the channel must be the
          first channel in the task. The only terminal you can use for
          E Series devices is PFI0.

        :param pattern:

          Specifies the digital pattern that must be met for the trigger to occur.

        :param match: Specifies if the conditions under which the trigger occurs

            'match' - Trigger when the signal matches the pattern

            'nomatch' - Trigger when the signal does NOT match the pattern

        :param pre_trigger_samps : uint32

          The minimum number of samples per channel to acquire before
          recognizing the Reference Trigger. The number of posttrigger
          samples per channel is equal to number of samples per channel
          in the NI-DAQmx Timing functions minus pretriggerSamples.
        """

        if not source.startswith('/'): # source needs to start with a '/'
            source = '/' + source

        self.lib.CfgDigPatternRefTrig(self, source, pattern, match, pre_trigger_samps)

    @Action()
    def disable_reference_trigger(self):
        """
        Disables reference triggering for the measurement or generation.

        Returns
        -------

          success_status : bool
        """
        return self.lib.DisableRefTrig(self) == 0


    #TODO CHECK
    def set_buffer(self, samples_per_channel):
        """
        Overrides the automatic I/O buffer allocation that NI-DAQmx performs.

        Parameters
        ----------

        samples_per_channel : int

          The number of samples the buffer can hold for each channel
          in the task. Zero indicates no buffer should be
          allocated. Use a buffer size of 0 to perform a
          hardware-timed operation without using a buffer.

        Returns
        -------

          success_status : bool
        """
        #channel_io_type = self.channel_io_type
        #return CALL('Cfg%sBuffer' % (channel_io_type.title()), self, uInt32(samples_per_channel)) == 0
        pass

    @Feat(units='Hz')
    def sample_clock_rate(self):
        """Sample clock rate.

        Set to None to reset.
        """

        err, value = self.lib.GetSampClkRate(self, RetValue('f64'))
        return value

    @sample_clock_rate.setter
    def sample_clock_rate(self, value):
        if value is None:
            self.lib.ResetSampClkRate()
        else:
            self.lib.SetSampClkRate(value)

    @Feat()
    def convert_clock_rate(self):
        """Convert clock rate.

        The rate at which to clock the analog-to-digital
        converter. This clock is specific to the analog input section
        of multiplexed devices.

        By default, NI-DAQmx selects the maximum convert rate
        supported by the device, plus 10 microseconds per channel
        settling time. Other task settings, such as high channel
        counts or setting Delay, can result in a faster default
        convert rate.

        Set to None to reset
        """
        err, value = self.lib.GetAIConvRate(RetValue('f64'))
        return value

    @convert_clock_rate.setter
    def convert_clock_rate(self, value):
        if value is None:
            self.lib.ResetAIConvRate()
        else:
            self.lib.SetAIConvRate(value)

    def sample_clock_max_rate(self):
        """Maximum Sample Clock rate supported by the task,
        based on other timing settings. For output tasks, the maximum
        Sample Clock rate is the maximum rate of the DAC. For input
        tasks, NI-DAQmx calculates the maximum sampling rate
        differently for multiplexed devices than simultaneous sampling
        devices.

        For multiplexed devices, NI-DAQmx calculates the maximum
        sample clock rate based on the maximum AI Convert Clock rate
        unless you set Rate. If you set that property, NI-DAQmx
        calculates the maximum sample clock rate based on that
        setting. Use Maximum Rate to query the maximum AI Convert
        Clock rate. NI-DAQmx also uses the minimum sample clock delay
        to calculate the maximum sample clock rate unless you set
        Delay.

        For simultaneous sampling devices, the maximum Sample Clock
        rate is the maximum rate of the ADC.
        """
        err, value = self.lib.GetSampClkMaxRate(RetValue('f64'))
        return value

    # Not implemented:
    # DAQmxReadBinary*, DAQmxReadCounter*, DAQmxReadDigital*
    # DAQmxGetNthTaskReadChannel, DAQmxReadRaw
    # DAQmxWrite*
    # DAQmxExportSignal
    # DAQmxCalculateReversePolyCoeff, DAQmxCreateLinScale
    # DAQmxWaitForNextSampleClock
    # DAQmxSwitch*
    # DAQmxConnectTerms, DAQmxDisconnectTerms, DAQmxTristateOutputTerm
    # DAQmxResetDevice
    # DAQmxControlWatchdog*

    # DAQmxAOSeriesCalAdjust, DAQmxESeriesCalAdjust, DAQmxGet*,
    # DAQmxMSeriesCalAdjust, DAQmxPerformBridgeOffsetNullingCal, DAQmxRestoreLastExtCalConst
    # DAQmxSelfCal, DAQmxSetAIChanCalCalDate, DAQmxSetAIChanCalExpDate, DAQmxSSeriesCalAdjust
    # External Calibration, DSA Calibration, PXI-42xx Calibration, SCXI Calibration
    # Storage, TEDS
    # DAQmxSetAnalogPowerUpStates, DAQmxSetDigitalPowerUpStates
    # DAQmxGetExtendedErrorInfo

    @Feat(values={True: Constants.Val_AllowRegen, False: Constants.Val_DoNotAllowRegen})
    def regeneration_enabled(self):
        """Generating the same data more than once is allowed.

        Set to None to reset.
        """
        err, value = self.lib.GetWriteRegenMode(RetValue('i32'))
        return value

    @regeneration_enabled.setter
    def regeneration_enabled(self, value):
        if value is None:
            self.lib.ResetWriteRegenMode()
        else:
            self.lib.SetWriteRegenMode(value)

    #TODO CHECK
    @Feat(values={'digital_edge': Constants.Val_DigEdge, 'disable': Constants.Val_None, None: None})
    def arm_start_trigger_type(self):
        """the type of trigger to use to arm the task for a
        Start Trigger. If you configure an Arm Start Trigger, the task
        does not respond to a Start Trigger until the device receives
        the Arm Start Trigger.

        Use None to reset
        """

        err, value = self.lib.GetArmStartTrigType(RetValue('i32'))
        return value

    @arm_start_trigger_type.setter
    def arm_start_trigger_type(self, trigger_type):
        if trigger_type is None:
            self.lib.ResetArmStartTrigType()
        else:
            self.lib.SetArmStartTrigType(trigger_type)


    @Feat()
    def arm_start_trigger_source(self):
        """Rhe name of a terminal where there is a digital
        signal to use as the source of the Arm Start Trigger

        Use None to Reset
        """
        err, value = self.lib.GetDigEdgeArmStartTrigSrc(RetStr(default_buf_size))
        return value

    @arm_start_trigger_source.setter
    @arm_start_trigger_source.setter
    def arm_start_trigger_source(self, source):
        source = str (source)
        if source is None:
            self.lib.ResetDigEdgeArmStartTrigSrc()
        else:
            self.lib.SetDigEdgeArmStartTrigSrc(source)

    @Feat(values={None: None}.update(_EDGE_TYPES))
    def arm_start_trigger_edge(self):
        """on which edge of a digital signal to arm the task
        for a Start Trigger

        Set to None to reset
        """
        err, value = self.lib.GetDigEdgeArmStartTrigEdge(RetValue('i32'))

    @arm_start_trigger_edge.setter
    def arm_start_trigger_edge(self, edge):
        if edge is None:
            self.lib.ResetDigEdgeArmStartTrigEdge()
        else:
            self.lib.SetDigEdgeArmStartTrigEdge(edge)

    @Feat(values={None: None}.update(_TRIGGER_TYPES))
    def pause_trigger_type(self):
        """The type of trigger to use to pause a task.

        Set to None to Reset
        """
        err, value = self.lib.GetPauseTrigType(RetValue('i32'))
        return value

    @pause_trigger_type.setter
    def pause_trigger_type(self, trigger_type):
        if trigger_type is None:
            self.lib.ResetPauseTrigType()
        else:
            self.lib.SetPauseTrigType(trigger_type)

    @Feat()
    def pause_trigger_source(self):
        """The name of a virtual channel or terminal where
        there is an analog signal to use as the source of the trigger.

        For E Series devices, if you use a channel name, the channel
        must be the only channel in the task. The only terminal you
        can use for E Series devices is PFI0.
        """

        type = self.pause_trigger_type
        if type == 'digital_level':
            fun = self.lib.GetDigLvlPauseTrigSrc
        elif type == 'analog_level':
            fun = self.lib.GetAnlgLvlPauseTrigSrc
        elif type == 'analog_window':
            fun = self.lib.GetAnlgWinPauseTrigSrc
        else:
            raise errors.InstrumentError('Pause trigger type is not specified')

        err, value = fun(*RetStr(default_buf_size))
        return value

    @pause_trigger_source.setter
    def pause_trigger_source(self, source):

        type = self.pause_trigger_type
        if type == 'digital_level':
            fun = self.lib.SetDigLvlPauseTrigSrc
        elif type == 'analog_level':
            fun = self.lib.SetAnlgLvlPauseTrigSrc
        elif type == 'analog_window':
            fun = self.lib.SetAnlgWinPauseTrigSrc
        else:
            raise errors.InstrumentError('Pause trigger type is not specified')

        fun(source)

    @Feat()
    def pause_trigger_when(self):
        """
        Specifies whether the task pauses above or below the threshold
        you specify with Level.

        Specifies whether the task pauses while the trigger signal is
        inside or outside the window you specify with Bottom and Top.

        Specifies whether the task pauses while the signal is high or
        low.

        Set To None to reset
        """
        type = self.pause_trigger_type
        if type == 'digital_level':
            fun = self.lib.SetDigLvlPauseTrigWhen
            convert = _WHEN_TRIGGER_DIG
        elif type == 'analog_level':
            fun = self.lib.SetAnlgLvlPauseTrigWhen
            convert = _WHEN_TRIGGER_ALVL
        elif type == 'analog_window':
            fun = self.lib.SetAnlgWinPauseTrigWhen
            convert = _WHEN_TRIGGER_AWIN
        else:
            raise errors.InstrumentError('Pause trigger type is not specified')

        err, val = fun(RetValue('i32'))
        for key, value in convert.items():
            if key == val:
                return val
        else:
            raise ValueError(val)

    @pause_trigger_when.setter
    def pause_trigger_when (self, when=None):

        if when is None:
            self.lib.ResetDigLvlPauseTrigWhen()
            return

        type = self.pause_trigger_type
        if type == 'digital_level':
            fun = self.lib.SetDigLvlPauseTrigWhen
            convert = _WHEN_TRIGGER_DIG
        elif type == 'analog_level':
            fun = self.lib.SetAnlgLvlPauseTrigWhen
            convert = _WHEN_TRIGGER_ALVL
        elif type == 'analog_window':
            fun = self.lib.SetAnlgWinPauseTrigWhen
            convert = _WHEN_TRIGGER_AWIN
        else:
            raise errors.InstrumentError('Pause trigger type is not specified')

        fun(convert[when])

    def read_current_position (self):
        """Samples per channel the current position in the buffer.
        """
        err, value = self.lib.GetReadCurrReadPos(*RetValue('u64'))
        return value

    def samples_per_channel_available(self):
        """The number of samples available to read per channel.

         This value is the same for all channels in the task.
        """
        err, value = self.lib.GetReadAvailSampPerChan(*RetValue('u32'))
        return value

    def samples_per_channel_acquired(self):
        """The total number of samples acquired by each channel.

        NI-DAQmx returns a single value because this value is
        the same for all channels.
        """
        err, value = self.lib.GetReadTotalSampPerChanAcquired(*RetValue('u32'))
        return value

    @Action(units='seconds')
    def wait_until_done(self, timeout=-1):
        """Wait for the measurement or generation to complete. Use this
        function to ensure that the specified operation is complete
        before you stop the task.

        :param timeout: The maximum amount of time, in seconds, to wait for the
          measurement or generation to complete. The function returns
          an error if the time elapses before the measurement or
          generation is complete.

          A value of -1 (Constants.Val_WaitInfinitely) means to wait
          indefinitely.

          If you set timeout to 0, the function checks once and
          returns an error if the measurement or generation is not
          done.
        """
        if timeout < 0:
            timeout = Constants.Val_WaitInfinitely
        return self.lib.WaitUntilTaskDone(timeout)


class Channel(_Base):
    """A virtual channel is a collection of settings such as a name,
    a physical channel, input terminal connections, the type of measurement
    or generation, and can include scaling information.
    """

    IO_TYPE = None

    def __init__(self, task=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._task = None
        if task == 'create':
            task = Task.typed_task(self.IO_TYPE)()
            print(task)
        self.task = task

    @property
    def task(self):
        return self._task

    @task.setter
    def task(self, value):
        if not self._task is None:
            raise Exception
        self._task = value
        if value is not None:
            self.log_debug('Creating channel {} with {}'.format(self.CREATE_FUN, self._create_args))
            value.execute_fun(self.CREATE_FUN, *self._create_args)

    def _preprocess_args(self, name, *args):
        """Injects device_name to all call to the library
        """
        return super()._preprocess_args(name, *((self.task.task_handle, self.name) + args))

    @Feat(read_once=True, values=_CHANNEL_TYPES)
    def io_type(self):
        """The type of the virtual channel.
        """
        err, value = self.lib.GetChanType(RetValue('i32'))
        return value

    @Feat(read_once=True)
    def physical_channel_name(self,):
        """Name of the physical channel upon which this virtual channel is based.
        """
        err, value = self.lib.GetPhysicalChanName(*RetStr(default_buf_size))
        return value

    def operation_direction(self):
        return 'input' if self.CHANNEL_TYPE else 'output'

    @Feat(read_once=True)
    def is_global(self):
        """Indicates whether the channel is a global channel.
        """
        err, value = self.lib.GetChanIsGlobal(RetValue('u32'))
        return value

    # TODO: DAQmx*ChanDescr

    @Feat()
    def buffer_size_on_board(self):
        """The number of samples the I/O buffer can hold for each
        channel in the task.

        If on_board is True then specifies in samples per channel the
        size of the onboard I/O buffer of the device.

        See also
        --------
        set_buffer_size, reset_buffer_size
        """
        fun = self.__get_fun('GetBuf{}OnbrdBufSize')
        err, value = fun(RetValue('u32'))
        return value

    @Feat()
    def buffer_size(self):
        """The number of samples the I/O buffer can hold for each
        channel in the task.

        Set to 0 to perform a hardware-timed operation without
        using a buffer. Setting this property overrides the automatic I/O
        buffer allocation that NI-DAQmx performs.

        Set to None to reset.
        """
        fun = self.__get_fun('GetBuf{}BufSize')
        err, value = fun(RetValue('u32'))
        return value

    @buffer_size.setter
    def buffer_size(self, size):
        if size is None:
            fun = self.__get_fun('ResetBuf{}BufSize')
            err = fun()
        else:
            fun = self.__get_fun('SetBuf{}BufSize')
            err = fun(size)

    @Feat()
    def max(self):
        """Maximum value value.

        Set to None to reset.
        """
        fun = self.__get_fun('Get{}Max')
        err, value = fun(RetValue('f64'))
        return value

    @max.setter
    def max(self, value):
        if value is None:
            fun = self.__get_fun('Reset{}Max')
            err = fun()
        else:
            fun = self.__get_fun('Set{}Max')
            err = fun(value)

    @Feat()
    def min(self):
        """Minimum value value.

        Set to None to reset.
        """
        fun = self.__get_fun('Get{}Min')
        err, value = fun(RetValue('f64'))
        return value

    @min.setter
    def min(self, value):
        if value is None:
            fun = self.__get_fun('Reset{}Min')
            err = fun()
        else:
            fun = self.__get_fun('Set{}Min')
            err = fun(value)

    @Feat()
    def range_high(self):
        """The upper limit of the input range of the
        device. This value is in the native units of the device. On E
        Series devices, for example, the native units is volts.

        Set to None to reset.
        """
        fun = self.__get_fun('Get{}RngHigh')
        err, value = fun(RetValue('f64'))
        return value

    @range_high.setter
    def range_high(self, value):
        if value is None:
            fun = self.__get_fun('Reset{}RngHigh')
            err = fun()
        else:
            fun = self.__get_fun('Set{}RngHigh')
            err = fun(value)

    @Feat()
    def range_low(self):
        """The lower limit of the input range of the
        device. This value is in the native units of the device. On E
        Series devices, for example, the native units is volts.

        Set to None to reset.
        """
        fun = self.__get_fun('Get{}RngLow')
        err, value = fun(RetValue('f64'))
        return value

    @range_low.setter
    def range_low(self, value):
        if value is None:
            fun = self.__get_fun('Reset{}RngLow')
            err = fun()
        else:
            fun = self.__get_fun('Set{}RngLow')
            err = fun(value)

    @Feat()
    def gain(self):
        """Gain factor to apply to the channel.

        Set to None to reset.
        """
        fun = self.__get_fun('Get{}Gain')
        err, value = fun(RetValue('f64'))
        return value

    @gain.setter
    def gain(self, value):
        if value is None:
            fun = self.__get_fun('Reset{}Gain')
            err = fun()
        else:
            fun = self.__get_fun('Set{}Gain')
            err = fun(value)

    @Feat()
    def number_of_lines(self):
        """Number of digital lines in the channel.
        """
        err, value = self.__get_fun('Get{}NumLines')(RetValue('u32'))
        return value


    def get_units(self, channel_name):
        """
        Specifies in what units to generate voltage on the
        channel. Write data to the channel in the units you select.

        Specifies in what units to generate current on the
        channel. Write data to the channel is in the units you select.

        See also
        --------
        set_units, reset_units
        """
        channel_name = str(channel_name)
        mt = self.get_measurment_type(channel_name)
        channel_type = self.channel_type
        if mt=='voltage':
            d = int32(0)
            CALL('Get%sVoltageUnits' % (channel_type), self, channel_name, ctypes.byref(d))
            units_map = {Contants.Val_Volts:'volts',
                         #Constants.Val_FromCustomScale:'custom_scale',
                         #Constants.Val_FromTEDS:'teds',
            }
            return units_map[d.value]
        raise NotImplementedError('{} {}'.format(channel_name, mt))


    @Feat(values={'none': Constants.Val_None, 'once': Constants.Val_Once,
                  'every_sample': Constants.Val_EverySample})
    def auto_zero_mode(self):
        """When to measure ground. NI-DAQmx subtracts the
        measured ground voltage from every sample.

        Set to None to reset.
        """
        fun = self.__get_fun('Get{}AutoZeroMode')
        err, value = fun(RetValue('i32'))
        return value

    @auto_zero_mode.setter
    def auto_zero_mode(self, value):
        if value is None:
            fun = self.__get_fun('Reset{}AutoZeroMode')
            err = fun()
        else:
            fun = self.__get_fun('Set{}AutoZeroMode')
            err = fun(value)

    @Feat(values={'dma': Constants.Val_DMA, 'interrupts': Constants.Val_Interrupts,
                  'programmed_io': Constants.Val_ProgrammedIO,
                  'usb': Constants.Val_USBbulk})
    def data_transfer_mode(self):
        """When to measure ground. NI-DAQmx subtracts the
        measured ground voltage from every sample.

        Set to None to reset.
        """
        fun = self.__get_fun('Get{}DataXferMech')
        err, value = fun(RetValue('i32'))
        return value

    @data_transfer_mode.setter
    def data_transfer_mode(self, value):
        if value is None:
            fun = self.__get_fun('Reset{}DataXferMech')
            err = fun()
        else:
            fun = self.__get_fun('Set{}DataXferMech')
            err = fun(value)

    @Feat(values={True: 1, False: 0, None: None})
    def duplicate_count_prevention_enabled(self):
        """Duplicate count prevention enabled.

        Set to None to Reset
        """
        err, value = self.lib.GetCIDupCountPrevent(RetValue('u32'))
        return value != 0

    @duplicate_count_prevention_enabled.setter
    def duplicate_count_prevention_enabled(self, value):
        if value is None:
            err = self.lib.ResetCIDupCountPrevent()
        else:
            err = self.lib.SetCIDupCountPrevent(value)

    @Feat()
    def timebase_rate(self):
        """Frequency of the counter timebase (Hz)

        TODO Can I put units and still None

        Set to None to reset.
        See also
        --------
        set_timebase_rate, reset_timebase_rate
        """
        err, value = self.lib.GetCICtrTimebaseRate(RetValue('f64'))
        return value

    @timebase_rate.setter
    def timebase_rate(self, value):
        if value is None:
            err = self.lib.ResetCICtrTimebaseRate()
        else:
            err = self.lib.SetCICtrTimebaseRate(value)


    @Feat(None)
    def terminal_pulse (self, terminal):
        """Terminal to generate pulses.
        """
        self.lib.SetCOPulseTerm(terminal)


    @Feat(None)
    def terminal_count_edges(self, channel, terminal):
        """Input terminal of the signal to measure.
        """
        self.lib.SetCICountEdgesTerm(terminal)
"""

DoneEventCallback_map = dict(AI=ctypes.CFUNCTYPE (int32, AnalogInputTask, int32, void_p),
                             AO=ctypes.CFUNCTYPE (int32, AnalogOutputTask, int32, void_p),
                             DI=ctypes.CFUNCTYPE (int32, DigitalInputTask, int32, void_p),
                             DO=ctypes.CFUNCTYPE (int32, DigitalOutputTask, int32, void_p),
                             CI=ctypes.CFUNCTYPE (int32, CounterInputTask, int32, void_p),
                             CO=ctypes.CFUNCTYPE (int32, CounterOutputTask, int32, void_p),
                             )
EveryNSamplesEventCallback_map = dict(AI=ctypes.CFUNCTYPE (int32, AnalogInputTask, int32, uInt32, void_p),
                                      AO=ctypes.CFUNCTYPE (int32, AnalogOutputTask, int32, uInt32, void_p),
                                      DI=ctypes.CFUNCTYPE (int32, DigitalInputTask, int32, uInt32, void_p),
                                      DO=ctypes.CFUNCTYPE (int32, DigitalOutputTask, int32, uInt32, void_p),
                                      CI=ctypes.CFUNCTYPE (int32, CounterInputTask, int32, uInt32, void_p),
                                      CO=ctypes.CFUNCTYPE (int32, CounterOutputTask, int32, uInt32, void_p),
                                      )
SignalEventCallback_map = dict(AI=ctypes.CFUNCTYPE (int32, AnalogInputTask, int32, void_p),
                               AO=ctypes.CFUNCTYPE (int32, AnalogOutputTask, int32, void_p),
                               DI=ctypes.CFUNCTYPE (int32, DigitalInputTask, int32, void_p),
                               DO=ctypes.CFUNCTYPE (int32, DigitalOutputTask, int32, void_p),
                               CI=ctypes.CFUNCTYPE (int32, CounterInputTask, int32, void_p),
                               CO=ctypes.CFUNCTYPE (int32, CounterOutputTask, int32, void_p),
                               )

"""

if __name__ == '__main__':

    import lantz.log

    lantz.log.log_to_screen(lantz.log.DEBUG)

    inst = System()
    print(inst.version)
    print(inst.device_names)



"""
    lantz.drivers.amplifier_research.arxx
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of 25S1G4A/AR30W1000B Microwave amplifier

    Author: Thomas Kiely
    Date: 6/20/2016
"""

import numpy as np
from lantz import Action, Feat, DictFeat, ureg
from lantz.messagebased import MessageBasedDriver

class ARxx(MessageBasedDriver):

        DEFAULTS = {
            'COMMON': {
                'write_termination': '\r\n',
                'read_termination': '\n',
            }
        }

        @Action()
        def power_on(self):
            self.write('P1')

        @Action()
        def power_off(self):
            self.write('P0')

        @Feat()
        def gain(self):
            """
            sets gain from 0 (min) to 4095 (max)
            this range encompasses a minimum of 10dB
            the maximum value of this range is at least 44dB
            """
            return int(self.query('G?')[1:])

        @gain.setter
        def gain(self,value):
            stringValue = '{0:0=4d}'.format(value)
            self.write('G{:s}'.format(stringValue))

        @Action()
        def reset(self):
            """
            resets device, clearing all faults, if possible
            """
            self.write('R')



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.network
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements a base class for drivers that communicate with instruments via TCP.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import socket

from lantz import Driver
from lantz.drivers.legacy.textual import TextualMixin
from lantz.errors import LantzTimeoutError


class LantzSocketTimeoutError(socket.timeout, LantzTimeoutError):
    pass


class TCPRawDriver(Driver):
    """Base class for drivers that communicate with instruments via TCP.

    :param host: Address of the network resource
    :param port: Port number
    """

    RECV_TERMINATION = '\n'
    SEND_TERMINATION = '\n'

    RECV_CHUNK = 1024

    def __init__(self, host='localhost', port=9997, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.host_port = (host, port)

    def raw_send(self, data):
        """Send raw bytes to the instrument.

        :param data: bytes to be sent to the instrument.
        :param data: bytes.
        """
        try:
            self.socket.send(data)
        except socket.timeout as e:
            raise LantzSocketTimeoutError(str(e))

    def raw_recv(self, size):
        """Receive raw bytes to the instrument.

        :param size: number of bytes to receive.
        :return: received bytes.
        :return type: bytes.
        """
        try:
            return self.socket.recv(size)
        except socket.timeout as e:
            raise LantzSocketTimeoutError(str(e))

    def initialize(self):
        self.log_debug('Opening port {}', self.host_port)
        return self.socket.connect(self.host_port)

    def finalize(self):
        self.log_debug('Closing port {}', self.host_port)
        return self.socket.close()

    def is_open(self):
        return self.socket.isOpen()


class TCPDriver(TCPRawDriver, TextualMixin):
    """Base class for drivers that communicate with instruments via TCP.

    :param host: Address of the network resource
    :param port: Port number
    """

    RECV_TERMINATION = '\n'
    SEND_TERMINATION = '\n'

    RECV_CHUNK = 1024




# -*- coding: utf-8 -*-
"""
    lantz.drivers.yokogawa
    ~~~~~~~~~~~~~~~~~~~~~

    :company: Yokogawa
    :description: Measurement instruments
    :website: http://www.yokogawa.com/

    ---
"""
from .yokogawa7651 import Yokogawa7651

__all__ = ['Yokogawa7651']



# -*- coding: utf-8 -*-
"""
    lantz.drivers.stanford.sr400
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of SR400 GATED PHOTON COUNTER

    Author: Alexandre Bourassa
    Date: 10/23/2016
"""

from lantz import Action, Feat, DictFeat, ureg
from lantz.messagebased import MessageBasedDriver

import time as _t

class SR400(MessageBasedDriver):
    DEFAULTS = {
        'COMMON': {
            'write_termination': '\r\n',
            'read_termination': '\r\n',
        }
    }

    COUNTERS = {'A':0, 'B':1, 'T':2}

    def write(self, *args, **kwargs):
        super().write(*args, **kwargs)
        _t.sleep(0.05)

    @Action()
    def reset(self):
        self.write('CL')
        _t.sleep(0.1)

    @Feat(values={'A,B FOR T PRESET': 0, 'A-B FOR T PRESET': 1, 'A+B FOR T PRESET': 2, 'A FOR B PRESET': 3})
    def counting_mode(self):
        return int(self.query('CM'))

    @counting_mode.setter
    def counting_mode(self, value):
        self.write('CM {}'.format(value))

    @DictFeat(keys=['A','B','T'], values={'10 MHz': 0, 'INPUT 1': 1, 'INPUT 2': 2, 'TRIG': 3})
    def counter_input(self, key):
        return int(self.query('CI {}'.format(self.COUNTERS[key])))

    @counter_input.setter
    def counter_input(self, key, value):
        if key is 'A' and not value in [0,1]: raise Exception("Counter A input can only be '10 MHz' or 'INPUT 1'")
        if key is 'B' and not value in [1,2]: raise Exception("Counter B input can only be 'INPUT 1' or 'INPUT 2'")
        if key is 'T' and not value in [0,2,3]: raise Exception("Counter T input can only be '10 MHz', 'INPUT 2' or 'TRIG'")
        self.write('CI {}, {}'.format(self.COUNTERS[key], value))

    @Feat(units='seconds')
    def dwell_time(self):
        return float(self.query('DT'))

    @dwell_time.setter
    def dwell_time(self, value):
        if not int(value) is 0:
            if not 2E-3<=value<=6E1:
                raise Exception("Dwell time must be either 0 (External) or between 2e-3 and 6e1 (only first significant digit matters)")
        self.write('DT {}'.format(value))

    @DictFeat(keys=['B','T'], limits=(1,9E11))
    def counter_preset(self, key):
        return int(float(self.query('CP {}'.format(self.COUNTERS[key]))))

    @counter_preset.setter
    def counter_preset(self, key, value):
        self.write('CP {}, {}'.format(self.COUNTERS[key], value))

    @Feat(limits=(1,2000))
    def period_per_scan(self):
        return int(self.query('NP'))

    @period_per_scan.setter
    def period_per_scan(self, value):
        self.write('NP {}'.format(value))

    @DictFeat(keys=['A', 'B'], values={'CW': 0, 'FIXED': 1, 'SCAN': 2})
    def gate_mode(self, key):
        return int(self.query('GM {}'.format(self.COUNTERS[key])))

    @gate_mode.setter
    def gate_mode(self, key, value):
        self.write('GM {}, {}'.format(self.COUNTERS[key], value))

    @DictFeat(keys=['A', 'B'], units='seconds', limits=(0,999.2E-3, 1e-9))
    def gate_delay(self, key):
        return float(self.query('GD {}'.format(self.COUNTERS[key])))

    @gate_delay.setter
    def gate_delay(self, key, value):
        self.write('GD {}, {}'.format(self.COUNTERS[key], value))

    @DictFeat(keys=['A', 'B'], units='seconds', limits=(0,99.92E-3, 1e-9))
    def scan_gate_delay(self, key):
        return float(self.query('GY {}'.format(self.COUNTERS[key])))

    @scan_gate_delay.setter
    def scan_gate_delay(self, key, value):
        self.write('GY {}, {}'.format(self.COUNTERS[key], value))

    @DictFeat(keys=['A', 'B'], units='seconds', limits=(0.005E-6, 99.92E-3, 1e-9))
    def gate_width(self, key):
        return float(self.query('GW {}'.format(self.COUNTERS[key])))

    @gate_width.setter
    def gate_width(self, key, value):
        self.write('GW {}, {}'.format(self.COUNTERS[key], value))

    @DictFeat(keys=['A', 'B', 'T'], values={'RISE': 0, 'FALL': 1})
    def disc_slope(self, key):
        return int(self.query('DS {}'.format(self.COUNTERS[key])))

    @disc_slope.setter
    def disc_slope(self, key, value):
        self.write('DS {}, {}'.format(self.COUNTERS[key], value))

    @DictFeat(keys=['A', 'B', 'T'], limits=(-0.3,0.3,0.0002), units='V')
    def disc_level(self, key):
        return float(self.query('DL {}'.format(self.COUNTERS[key])))

    @disc_level.setter
    def disc_level(self, key, value):
        self.write('DL {}, {}'.format(self.COUNTERS[key], value))



# -*- coding: utf-8 -*-
"""
    lantz.drivers.ni.daqmx
    ~~~~~~~~~~~~~~~~~~~~~~

    Implements bindings to the DAQmx (windows) National Instruments libraries.

    Sources::

        - DAQmx Reference manual
        - DAQmx Base Reference manual
        - pylibnidaqmx
          http://pylibnidaqmx.googlecode.com


    :company: National Instruments
    :description: 
    :website: http://www.ni.com/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .base import System, Task, Channel, Device
from .channels import *
from .tasks import *
from .constants import Constants, Types





"""
    lantz.drivers.keithley.smu2400
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of the 2400 series Source Meter Unit

    Author: Alexandre Bourassa
    Date: 09/08/2016
"""


from lantz import Action, Feat, DictFeat, ureg
from lantz.messagebased import MessageBasedDriver
import time as _t

class QOSIM900(MessageBasedDriver):

        DEFAULTS = {
            'COMMON': {
                'write_termination': '\n',
                'read_termination': '\n',
            }
        }

        AMP_CHS = {0:3, 1:4, 2:5, 3:6, 4:7, 5:8}
        SIM922_ch = 1

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._buffered_adc_gain = None


        def write_to_card(self, channel, cmd):
            s = "SNDT {}, '{}'".format(channel, cmd)
            return self.write(s)

        def read_to_card(self, channel):
            n = int(self.query('NINP? {}'.format(channel)))
            ans = self.query("RAWN? {}, {}".format(channel, n))
            return ans.strip()

        def query_to_card(self, channel, cmd, delay=0.1):
            self.write_to_card(channel, cmd)
            _t.sleep(delay)
            return self.read_to_card(channel)

        @DictFeat(units='K', keys=[1,2])
        def temperature(self, ch):
            return float(self.query_to_card(self.SIM922_ch, 'TVAL? {}'.format(ch), delay=0.5))

        @DictFeat(keys=list(AMP_CHS.keys()), limits=(0, 25e-6, 25e-6/65535), units='A')
        def bias_current(self, key):
            return float(25e-6*int(self.query_to_card(self.AMP_CHS[key], '+B?'))/65535)*ureg.A

        @bias_current.setter
        def bias_current(self, key, val):
            return self.write_to_card(self.AMP_CHS[key], '+B{};'.format(int(65535*val/25e-6)))

        @DictFeat(keys=list(AMP_CHS.keys()),limits=(0,2.55,0.01), units='s')
        def reset_event_duration(self, key):
            return (int(self.query_to_card(self.AMP_CHS[key], '+D?'))/100.0)*ureg.s

        @reset_event_duration.setter
        def reset_event_duration(self, key, val):
            return self.write_to_card(self.AMP_CHS[key], '+D{};'.format(int(100*val)))

        @DictFeat(keys=list(AMP_CHS.keys()), values={True: '1', False: '0'})
        def autoreset_enable(self, key):
            return self.query_to_card(self.AMP_CHS[key], '+E?')

        @autoreset_enable.setter
        def autoreset_enable(self, key, val):
            return self.write_to_card(self.AMP_CHS[key], '+E{};'.format(val))

        @Action()
        def reset_bias(self, key):
            if not key in list(self.AMP_CHS.keys()): raise Exception('Specified <key> not in channel list.  Please check AMP_CH')
            return self.write_to_card(self.AMP_CHS[key], '+F;')

        @Action()
        def autobias(self, key):
            if not key in list(self.AMP_CHS.keys()): raise Exception('Specified <key> not in channel list.  Please check AMP_CH')
            return self.write_to_card(self.AMP_CHS[key], '+G;')

        @Action()
        def set_adc_gain(self, key, state='low'):
            if not key in list(self.AMP_CHS.keys()): raise Exception('Specified <key> not in channel list.  Please check AMP_CH')
            if not state in ['low','high']: raise Exception("<state> can only be 'low' or 'high'")
            self._buffered_adc_gain = state
            return self.write_to_card(self.AMP_CHS[key], '+C{};'.format(int(state=='high')))

        @DictFeat(keys=list(AMP_CHS.keys()), units='V')
        def adc_voltage(self, key):
            if not self._buffered_adc_gain in ['low','high']: raise Exception('adc_gain unknown, please set using <inst>.set_adc_gain(channel, state)')
            else:
                gfactor = (5.0 if self._buffered_adc_gain=='low' else 1.1)
                return float(gfactor*int(self.query_to_card(self.AMP_CHS[key], '+C?'))/65535)*ureg.V

        @DictFeat(keys=list(AMP_CHS.keys()), limits=(0, 25e-6, 25e-6/65535), units='A')
        def mem_bias_current(self, key):
            return float(25e-6*int(self.query_to_card(self.AMP_CHS[key], '+H?'))/65535)*ureg.A

        @mem_bias_current.setter
        def mem_bias_current(self, key, val):
            """DO NOT CALL THIS FUNCTION REPEATEDLY.  THE MEMORY AS A 100,000 WRITE LIFETIME
            """
            print('WARNING: DO NOT CALL THIS FUNCTION REPEATEDLY.  THE MEMORY AS A 100,000 WRITE LIFETIME')
            return self.write_to_card(self.AMP_CHS[key], '+H{};'.format(int(65535*val/25e-6)))

        @Action()
        def set_bias_to_mem(self, key):
            if not key in list(self.AMP_CHS.keys()): raise Exception('Specified <key> not in channel list.  Please check AMP_CH')            
            return self.write_to_card(self.AMP_CHS[key], '+H;')

        @DictFeat(keys=list(AMP_CHS.keys()) ,read_once=True)
        def card_idn(self, key):
            return self.query_to_card(self.AMP_CHS[key], '+A?')

        @Feat(read_once=True)
        def idn(self):
            return self.query('*IDN?')

        @Action()
        def reset(self):
            return self.write('*RST')



# -*- coding: utf-8 -*-
"""
    lantz.drivers.rgblasersystems.minilasevo
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from pyvisa import constants

from lantz import Feat
from lantz import MessageBasedDriver
from lantz import errors


class MiniLasEvo(MessageBasedDriver):
    """Driver for any RGB Lasersystems MiniLas Evo laser.
    """

    DEFAULTS = {'ASRL': {'write_termination': '\r\n',
                         'read_termination': '\r\n',
                         'baud_rate': 57600,
                         'bytesize': 8,
                         'parity': constants.Parity.none,
                         'stop_bits': constants.StopBits.one,
                         'encoding': 'ascii'
                        }}

    #: flow control flags
    #RTSCTS = False
    #DSRDTR = False
    #XONXOFF = False

    def query(self, command, *, send_args=(None, None),
              recv_args=(None, None)):
        """Send query to the laser and return the answer, after handling
        possible errors.

        :param command: command to be sent to the instrument
        :type command: string

        :param send_args: (termination, encoding) to override class defaults
        :param recv_args: (termination, encoding) to override class defaults
        """
        ans = super().query(command, send_args=send_args, recv_args=recv_args)
        # TODO: Echo handling
        code = ans[0]
        if code != 0:
            if code == '1':
                raise errors.InstrumentError('Command invalid')
            elif code == '2':
                raise errors.InstrumentError('Wrong number of parameters')
            elif code == '3':
                raise errors.InstrumentError('Parameter value is out of range')
            elif code == '4':
                raise errors.InstrumentError('Unlocking code is wrong')
            elif code == '5':
                raise errors.InstrumentError('Device is locked for this command')
            elif code == '6':
                raise errors.InstrumentError('This function is not supported')
            elif code == '7':
                raise errors.InstrumentError('Timeout while reading command (60 s)')
            elif code == '8':
                raise errors.InstrumentError('This value is currently not available')

        ans = ans[2:]
        # TODO: Code reporting?
        return ans

    @Feat(read_once=True)
    def idn(self):
        """Identification of the device
        """
        manufacturer = self.query('DM?')
        device = self.query('DT?')
        serial = self.query('DS?')
        ans = manufacturer + ', ' + device + ', serial number ' + serial
        return ans

    @Feat()
    def status(self):
        """Current device status
        """
        ans = self.query('S?')
        if ans == '0x10':
            ans = 'Temperature of laser head is ok'
        elif ans == '0x01':
            ans = 'Laser system is active, radiation can be emitted'
        elif ans == '0x02':
            ans = '(reserved)'
        elif ans == '0x04':
            ans = 'The interlock is open'
        elif ans == '0x08':
            ans = self.query('E?')
            if ans == '0x01':
                ans = 'Temperature of laser head is too high'
            elif ans == '0x02':
                ans = 'Temperature of laser head is too low'
            elif ans == '0x04':
                ans = 'Temperature-sensor connection is broken'
            elif ans == '0x08':
                ans = 'Temperature sensor cable is shortened'
            elif ans == '0x40':
                ans = 'Current for laser head is too high'
            elif ans == '0x80':
                ans = 'Internal error (laser system cannot be activated)'
        return ans

    @Feat()
    def operating_hours(self):
        """Total operating hours [hhhh:mm]
        """
        return self.query('R?')

    @Feat()
    def software_version(self):
        """Software version
        """
        return self.query('DO?')

    @Feat(read_once=True)
    def emission_wavelenght(self):
        """Emission wavelenght in nm
        """
        return self.query('DW?')

    @Feat()
    def available_features(self):
        """Available features (reserved for future use)
        """
        return self.query('DF?')

    @Feat()
    def control_mode(self):
        """Active current (power) control
        """
        ans = self.query('DC?')
        if ans == 'ACC':
            ans = 'Active current control'
        else:
            ans = 'Active power control'
        return ans

    # TEMPERATURE

    @Feat()
    def temperature(self):
        """Current temperature in ºC
        """
        return self.query('T?')

    @Feat(read_once=True)
    def temperature_min(self):
        """Lowest operating temperature in ºC
        """
        return self.query('LTN?')

    @Feat(read_once=True)
    def temperature_max(self):
        """Highest operating temperature in ºC
        """
        return self.query('LTP?')

    # ENABLED REQUEST

    @Feat(values={True: '1', False: '0'})
    def enabled(self):
        """Method for turning on the laser
        """
        return self.query('O?')

    @enabled.setter
    def enabled(self, value):
        self.query('O=' + value)

    # LASER POWER

    def initialize(self):
        super().initialize()
        self.power = 0
        self.feats.power.limits = (0, self.maximum_power.magnitude)

    @Feat(units='mW')
    def maximum_power(self):
        """Gets the maximum emission power of the laser
        """
        return float(self.query('LP?'))

    @Feat(units='mW')
    def power_sp(self):
        """Gets and sets the emission power
        """
        return float(self.query('P?'))

    @power_sp.setter
    def power(self, value):
        self.query('P={:.1f}'.format(value))

if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')
    parser.add_argument('-p', '--port', type=str, default='17',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_screen(lantz.log.DEBUG)
    with MiniLasEvo(args.port) as inst:
        if args.interactive:
            from lantz.ui.app import start_test_app
            start_test_app(inst)
        else:
            # Add your test code here
            print('Non interactive mode')
            """inst.power = 0
            inst.temperature
            print(inst.power)
            inst.power = 0
            print(inst.idn)
            """




from lantz import Feat
from lantz.messagebased import MessageBasedDriver

class Lakeshore475(MessageBasedDriver):

    @Feat(units='gauss')
    def field(self):
        return self.query('RDGFIELD?')
    
    @Feat(units='Hz')
    def frequency(self):
        return self.query('RDGFRQ?')




"""
    lantz.drivers.keithley.smu2400
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of the 2400 series Source Meter Unit

    Author: Alexandre Bourassa
    Date: 09/08/2016
"""


from lantz import Action, Feat, DictFeat, ureg
from lantz.messagebased import MessageBasedDriver

class SMU2400(MessageBasedDriver):

        DEFAULTS = {
            'COMMON': {
                'write_termination': '\n',
                'read_termination': '\n',
            }
        }

        @Feat(read_once=True)
        def idn(self):
            return self.query('*IDN?')


        @Feat(values={'current': 'CURR:DC', 'voltage': 'VOLT:DC', 'resistance':'RES'})
        def meas_config(self):
            """
            Configures the instrument to a specific setup for measurements on the specified function:
            current, voltage or resistance

            WARNING: This may reset some configurations, and this will turn on the output
            """
            return self.query(':CONF?').strip('"').split('"')[-1]

        @meas_config.setter
        def meas_config(self, value):
            self.write(':CONF:{}'.format(value))


        @Feat(values={False:'0', True:'1'})
        def output(self):
            return self.query(':OUTP:STAT?')

        @output.setter
        def output(self, value):
            return self.write(':OUTP:STAT {}'.format(value))


        @Feat(units='A', limits=(-1.05, 1.05,))
        def current_compliance(self):
            return self.query(':CURR:PROT?')

        @current_compliance.setter
        def current_compliance(self, value):
            return self.write(':CURR:PROT {}'.format(value))


        @Feat(units='V', limits=(-210., 210.,))
        def voltage_compliance(self):
            return self.query(':VOLT:PROT?')

        @voltage_compliance.setter
        def voltage_compliance(self, value):
            return self.write(':VOLT:PROT {}'.format(value))


        @Feat(units='V', limits=(-210., 210.,))
        def sense_voltage_range(self):
            return self.query(':SENS:VOLT:RANG?')

        @sense_voltage_range.setter
        def sense_voltage_range(self, value):
            return self.write(':SENS:VOLT:RANG {}'.format(value))


        @Feat(units='A', limits=(-1.05, 1.05,))
        def sense_current_range(self):
            return self.query(':SENS:CURR:RANG?')

        @sense_current_range.setter
        def sense_current_range(self, value):
            return self.write(':SENS:CURR:RANG {}'.format(value))


        @Feat(units='V', limits=(-210., 210.,))
        def source_voltage_range(self):
            return self.query(':SOUR:VOLT:RANG?')

        @source_voltage_range.setter
        def source_voltage_range(self, value):
            return self.write(':SOUR:VOLT:RANG {}'.format(value))


        @Feat(units='A', limits=(-1.05, 1.05,))
        def source_current_range(self):
            return self.query(':SOUR:CURR:RANG?')

        @source_current_range.setter
        def source_current_range(self, value):
            return self.write(':SOUR:CURR:RANG {}'.format(value))


        @Feat(units='A', limits=(-1.05, 1.05,))
        def source_current(self):
            return self.query(':SOUR:CURR?')

        @source_current.setter
        def source_current(self, value):
            return self.write(':SOUR:CURR {}'.format(value))


        @Feat(units='V', limits=(-210., 210.,))
        def source_voltage(self):
            return self.query(':SOUR:VOLT?')

        @source_voltage.setter
        def source_voltage(self, value):
            return self.write(':SOUR:VOLT {}'.format(value))


        @Feat(values={'current':'CURR', 'voltage':'VOLT', 'memory':'MEM'})
        def source_function(self):
            return self.query(':SOUR:FUNC?')

        @source_function.setter
        def source_function(self, value):
            return self.write(':SOUR:FUNC {}'.format(value))


        SENSE_FUNCTION = {'current':'CURR:DC', 'voltage':'VOLT:DC', 'resistance':'RES'}
        @DictFeat(values={'ON':'1', 'OFF':'0'}, keys=SENSE_FUNCTION)
        def sense_function(self, key):
            return self.query(':FUNC:STAT? "{}"'.format(key))

        @sense_function.setter
        def sense_function(self, key, value):
            value = 'OFF' if value=='0' else 'ON'
            return self.write(':SENS:FUNC:{} "{}"'.format(value, key))


        @Feat(values={'OFF': '0', 'ON': '1'})
        def sense_concurent(self):
            return self.query(':FUNC:CONC?')

        @sense_concurent.setter
        def sense_concurent(self, value):
            return self.write(':FUNC:CONC {}'.format(value))



        @Action()
        def reset(self):
            return self.write('*RST')

        @Action()
        def get_data(self):
            """
            Perform a read operation that returns a dictionary
            """
            ans = list(map(float, self.query(':READ?').split(',')))
            return {'voltage': ans[0], 'current': ans[1], 'resistance': ans[2], 'time': ans[3], 'status': ans[4]}



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.aeroflex.a2023a
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers for an signal generator.


    Sources::

        - Aeroflex 2023a Manual.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz import Feat, Action
from lantz.drivers.legacy.serial import SerialDriver


class A2023a(SerialDriver):
    """Aeroflex Test Solutions 2023A 9 kHz to 1.2 GHz Signal Generator.
    """

    RECV_TERMINATION = 256
    SEND_TERMINATION = '\n'

    @Feat(read_once=True)
    def idn(self):
        """Instrument identification.
        """
        return self.parse_query('*IDN?',
                                format='{manufacturer:s},{model:s},{serialno:s},{softno:s}')

    @Feat(read_once=True)
    def fitted_options(self):
        """Fitted options.
        """
        return self.query('*OPT?').split(',')

    @Action()
    def reset(self):
        """Set the instrument functions to the factory default power up state.
        """
        self.send('*RST')

    @Action()
    def self_test(self):
        """Is the interface and processor are operating?
        """
        return self.query('*TST?') == '0'

    @Action()
    def wait(self):
        """Inhibit execution of an overlapped command until the execution of
        the preceding operation has been completed.
        """
        self.send('*WAI')

    @Action()
    def trigger(self):
        """Equivalent to Group Execute Trigger.
        """
        self.send('*TRG')

    @Feat()
    def status_byte(self):
        """Status byte, a number between 0-255.
        """
        return int(self.query('*STB?'))

    @Feat()
    def service_request_enabled(self):
        """Service request enable register.
        """
        return int(self.query('*SRE?'))

    @service_request_enabled.setter
    def service_request_enabled(self, value):
        self.query('*SRE {0:d}'.format(value))

    @Feat()
    def event_status_reg(self):
        """Standard event enable register.
        """
        return int(self.query('*ESR?'))

    @event_status_reg.setter
    def event_status_reg(self, value):
        self.query('*ESR {0:d}'.format(value))

    @Feat()
    def event_status_enabled(self):
        """Standard event enable register.
        """
        return int(self.query('*ESR?'))

    @event_status_enabled.setter
    def event_status_enabled(self, value):
        self.query('*ESR {0:d}'.format(value))

    @Action()
    def clear_status(self):
        self.send('*CLS')

    @Feat(units='Hz')
    def frequency(self):
        """Carrier frequency.
        """
        return self.parse_query('CFRQ?',
                                format=':CFRQ:VALUE {0:f};{_}')

    @frequency.setter
    def frequency(self, value):
        self.send('CFRQ:VALUE {0:f}HZ'.format(value))


    @Feat(units='V')
    def amplitude(self):
        """RF amplitude.
        """
        return self.parse_query('RFLV?',
                                format=':RFLV:UNITS {_};TYPE {_};VALUE {0:f};INC {_};<status>')

    @amplitude.setter
    def amplitude(self, value):
        self.query('RFLV:VALUE {0:f}V'.format(value))


    @Feat(units='V')
    def offset(self):
        """Offset amplitude.
        """
        return self.parse_query('RFLV:OFFS?',
                                format=':RFLV:OFFS:VALUE {0:f};{_}')

    @offset.setter
    def offset(self, value):
        self.query('RFLV:OFFS:VALUE {0:f}'.format(value))

    @Feat(values={True: 'ENABLED', False: 'DISABLED'})
    def output_enabled(self):
        """Enable or disable the RF output
        """
        return self.query('OUTPUT?')

    @output_enabled.setter
    def output_enabled(self, value):
        self.query('OUTPUT:{}'.format(value))

    @Feat(units='deg')
    def phase(self):
        """Phase offset
        """
        return self.parse_query('CFRQ?',
                                format=':CFRQ:VALUE {:f}; INC {_};MODE {_}')

    @phase.setter
    def phase(self, value):
        """Adjust phase offset of carrier in degrees.
        """
        self.send('CFRQ:PHASE {}'.format(value))

    @Feat(values={'INT', 'EXT10DIR', 'EXTIND', 'EXT10IND', 'INT10OUT'})
    def frequency_standard(self):
        """Get internal or external frequency standard.
        """
        return self.query('FSTD?')

    @frequency_standard.setter
    def frequency_standard(self, value):
        self.query('FSTD {}'.format(value))

    @Feat()
    def rflimit(self):
        """Set RF output level max.
        """
        return self.query('*RFLV:LIMIT?')

    @rflimit.setter
    def rflimit(self, value):
        self.query('RFLV:LIMIT {}'.format(value))

    @Feat(values={True: 'ENABLED', False: 'DISABLED'})
    def rflimit_enabled(self):
        return self.query('*RFLV:LIMIT?')

    @rflimit_enabled.setter
    def rflimit_enabled(self, value):
        self.query('RFLV:LIMIT:{}'.format(value))

    def remote(self, value):
        if value:
            self.send('^A')
        else:
            self.send('^D')

    @Action(units='ms')
    def expose(self, exposure_time=1):
        self.send('EXPOSE {}'.format(exposure_time))

    @Feat(values={True: 'on', False: 'off'})
    def time(self):
        #TODO: Command??
        self.send('')
        return self.recv()

    @time.setter
    def time(self, value):
        self.send("vlal ".format(value))

    def local_lockout(self, value):
        if value:
            self.send('^R')
        else:
            self.send('^P')

    def software_handshake(self, value):
        if value:
            self.send('^Q')
        else:
            self.send('^S')


if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')
    parser.add_argument('-p', '--port', type=str, default='17',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_socket(lantz.log.DEBUG)
    with A2023a(args.port) as inst:
        if args.interactive:
            from lantz.ui.app import start_test_app
            start_test_app(inst)
        else:
            print(inst.idn)
            inst.fstd = "EXT10DIR"
            print(inst.fstd)
            print(inst.freq)
            inst.freq = 41.006
            print(inst.rflevel)
            inst.rflevel = -13
            inst.phase=0
            print(inst.phase)
            inst.phase=30
            print(inst.phase)
            inst.phase=60
            print(inst.phase)




# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy
    ~~~~~~~~~~~~~~~~~~~~

    Legacy drivers based on old instrument classes. Do not use for new code.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""




# -*- coding: utf-8 -*-
"""
    lantz.drivers.rigol
    ~~~~~~~~~~~~~~~~~~~

    :company: Rigol.
    :description: Acquisition devices
    :website: http://www.rigolna.com/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .ds1052e import DS1052e

__all__ = ['DS1052e']
